import unittest

import numpy as np
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem
from skmixed.lme.trees import Tree, Forest


class TestForest(unittest.TestCase):

    def test_forest(self):
        # TODO: reimplement fit_problem and predict_problem to make it work
        self.assertTrue(True)
        return None
        max_mse = 0.05
        min_explained_variance = 0.9
        categorical_effects_min_accuracy = 0.7
        active_categorical_features = (0, 2, 3)
        for i in range(1):  # TODO: make a better test for this functionality
            problem, true_parameters = LinearLMEProblem.generate(groups_sizes=[40, 30, 50],
                                                                 features_labels=[3, 6, 5, 5, 6],
                                                                 random_intercept=True,
                                                                 obs_std=0.1,
                                                                 seed=i)
            X, y = problem.to_x_y()
            continuous_model = LinearLMESparseModel(lb=0, lg=0, nnz_tbeta=2, nnz_tgamma=2)
            forest = Forest(continuous_model, num_trees=10, max_depth=3)
            forest.fit_problem(problem)
            y_pred = forest.predict_problem(problem)
            x, y_true = problem.to_x_y()
            lower_quantile, higher_quantile = forest.get_prediction_uncertainty(problem)
            errors = np.array([y_pred - lower_quantile, higher_quantile - y_pred])


from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem

import unittest

import numpy as np
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem


class TestLinearLMESparseModel_with_selectiveness(unittest.TestCase):

    def test_solving_sparse_problem(self):
        trials = 10
        problem_parameters = {
            "groups_sizes": [20, 12, 14, 50, 11],
            "features_labels": [3, 3, 3, 3, 3, 3, 3, 3, 3, 3],
            "random_intercept": True,
            "obs_std": 0.1,
        }

        model_parameters = {
            "lb": 0.01,
            "lg": 0.01,
            "initializer": None,
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
            "n_iter_outer": 20,
        }

        max_mse = 0.1
        min_explained_variance = 0.9
        fixed_effects_min_accuracy = 0.7
        random_effects_min_accuracy = 0.7

        fea = []
        rea = []

        for i in range(trials):
            np.random.seed(i)
            true_beta = np.random.choice(2, size=11, p=np.array([0.5, 0.5]))
            if sum(true_beta) == 0:
                true_beta[0] = 1
            true_gamma = np.random.choice(2, size=11, p=np.array([0.3, 0.7])) * true_beta

            problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters,
                                                                       beta=true_beta,
                                                                       gamma=true_gamma,
                                                                       seed=i)
            model = LinearLMESparseModel(**model_parameters,
                                         nnz_tbeta=sum(true_beta),
                                         nnz_tgamma=sum(true_gamma),
                                         regularization_type="loss-weighted"
                                         )
            model2 = LinearLMESparseModel(**model_parameters,
                                          nnz_tbeta=sum(true_beta),
                                          nnz_tgamma=sum(true_gamma),
                                          regularization_type="l2"
                                          )

            x, y = problem.to_x_y()
            model.fit_problem(problem)
            model2.fit_problem(problem)

            logger = model.logger_

            # TODO: It won't decrease monotonically because it may jump when we increase regularization.
            # loss = np.array(logger.get("loss"))
            # self.assertTrue(np.all(loss[1:] - loss[:-1] <= 0),
            #                 msg="%d) Loss does not decrease monotonically with iterations. (seed=%d)" % (i, i))

            y_pred = model.predict_problem(problem)
            explained_variance = explained_variance_score(y, y_pred)
            mse = mean_squared_error(y, y_pred)

            y_pred2 = model2.predict_problem(problem)
            explained_variance2 = explained_variance_score(y, y_pred2)
            mse2 = mean_squared_error(y, y_pred2)

            coefficients = model.coef_
            maybe_tbeta = coefficients["tbeta"]
            maybe_tgamma = coefficients["tgamma"]
            fixed_effects_accuracy = accuracy_score(true_beta, maybe_tbeta != 0)
            random_effects_accuracy = accuracy_score(true_gamma, maybe_tgamma != 0)

            coefficients2 = model2.coef_
            maybe_tbeta2 = coefficients2["tbeta"]
            maybe_tgamma2 = coefficients2["tgamma"]
            fixed_effects_accuracy2 = accuracy_score(true_beta, maybe_tbeta2 != 0)
            random_effects_accuracy2 = accuracy_score(true_gamma, maybe_tgamma2 != 0)
            # print("\n %d) MSE    EV FEA REA")
            # print("%.4f  %.4f %.4f %.4f" % (mse, explained_variance, fixed_effects_accuracy, random_effects_accuracy))
            # print("%.4f  %.4f %.4f %.4f" % (mse2, explained_variance2, fixed_effects_accuracy2, random_effects_accuracy2))

            # maybe_per_group_coefficients = coefficients["per_group_coefficients"]

            self.assertGreater(explained_variance, min_explained_variance,
                               msg="%d) Explained variance is too small: %.3f < %.3f. (seed=%d)"
                                   % (i,
                                      explained_variance,
                                      min_explained_variance,
                                      i))
            self.assertGreater(max_mse, mse,
                               msg="%d) MSE is too big: %.3f > %.2f  (seed=%d)"
                                   % (i,
                                      mse,
                                      max_mse,
                                      i))
            self.assertGreater(fixed_effects_accuracy, fixed_effects_min_accuracy,
                               msg="%d) Fixed Effects Selection Accuracy is too small: %.3f < %.2f  (seed=%d)"
                                   % (i,
                                      fixed_effects_accuracy,
                                      fixed_effects_min_accuracy,
                                      i)
                               )
            self.assertGreater(random_effects_accuracy, random_effects_min_accuracy,
                               msg="%d) Random Effects Selection Accuracy is too small: %.3f < %.2f  (seed=%d)"
                                   % (i,
                                      random_effects_accuracy,
                                      random_effects_min_accuracy,
                                      i)
                               )
            fea.append(fixed_effects_accuracy)
            rea.append(random_effects_accuracy)

        return None



import unittest

import numpy as np
import pandas as pd
from pathlib import Path
from numpy.linalg import lstsq
import json

from skmixed.lme.problems import LinearLMEProblem
from skmixed.lme.cascades import LMECascade


class TestLMECascade(unittest.TestCase):

    def test_solving_marlena_dataset(self):
        # This dataset was generated by this generator:
        # https://github.com/ihmeuw-msca/multilevel-model
        # The whole class appeared as a solution for a
        # small internal Math Sciences team competition in Sept. 2020.

        # Paths to dataset and true parameters/answers
        dataset = "simulation-2"
        dataset_generator_directory = Path("examples/cascades/datasets")
        dataset_path = dataset_generator_directory / dataset / (dataset + ".csv")
        hierarchy_path = dataset_generator_directory / "hierarchy.csv"
        true_coefficients_path = dataset_generator_directory / dataset / "KEY" / ("PARAMS-" + dataset + ".csv")
        true_answers_path = dataset_generator_directory / dataset / "KEY" / ("TRUTH-" + dataset + ".csv")
        true_parameters_path = dataset_generator_directory / dataset / ("SETTINGS-" + dataset + ".json")

        # Reading data
        data = pd.read_csv(dataset_path)
        hierarchy = pd.read_csv(hierarchy_path)
        hierarchy = hierarchy[hierarchy.location_id.isin(data.location_id.unique())]
        hierarchy = hierarchy[['location_name', 'location_id', 'parent_id', 'level']].reset_index(drop=True)
        data = data.merge(hierarchy, on="location_id", how="left")

        levels = sorted(data["level"].unique())

        covariates_colnames = [column for column in data.columns if column.startswith("x_")]

        data["intercept"] = 1
        data["predictions"] = 0
        data["residues"] = 0
        data["se"] = 0
        hierarchy["se"] = 1

        # Transform hierarchy into dict: {child_id: parent_id}
        hierarchy_dict = {row["location_id"]: row["parent_id"] for i, row in hierarchy.iterrows()}

        # Inferring observational standard error
        # NB: it's completely unnecessary to do it in this way:
        # you can (and should) estimate it along with beta and gamma.
        # However, this module was a hackaton-style work, so I did not have time
        # to implement it.
        current_level = data["level"] == 0
        level0_data = data[current_level][["intercept"] + covariates_colnames].to_numpy()
        level0_target = data[current_level]["observation"].to_numpy()
        level0_coefficients = lstsq(level0_data, level0_target, rcond=None)[0]
        data.loc[current_level, "predictions"] = level0_data.dot(level0_coefficients)
        data.loc[current_level, "residues"] = data.loc[current_level, "observation"] - data.loc[
            current_level, "predictions"]
        current_se = np.sqrt((data.loc[current_level, "residues"].var()))

        # Generating one problem per level
        problems = []
        for level in levels:
            current_level = data["level"] == level
            data.loc[current_level, "se"] = current_se
            X = data[current_level][covariates_colnames + ["location_id", "se"]].to_numpy()
            column_labels = [1] * len(covariates_colnames) + [0, 4]
            X = np.vstack([column_labels, X])
            y = data[current_level]["observation"].to_numpy()
            problem = LinearLMEProblem.from_x_y(X, y, random_intercept=True)
            problems.append(problem)

        # Fitting a model
        model = LMECascade()
        model.fit_problem(problems, hierarchy_dict)
        beta = model.coef_["beta"]

        # Aggregating answers into a dataframe
        answers = [{"location_id": group, "intercept": beta[0] + u[0], "effect_0": beta[1]} for group, u in
                   model.coef_["random_effects"].items()]
        coefficients = pd.DataFrame(answers)
        true_coefficients = pd.read_csv(true_coefficients_path)
        coefficients = coefficients.merge(true_coefficients, on="location_id", how="inner", suffixes=('_pred', '_true'))
        with open(true_parameters_path) as f:
            true_parameters = json.load(f)

        # In case you want to see the results
        # print(coefficients[["location_id", "intercept_true", "intercept_pred", "effect_0_true", "effect_0_pred"]])

        # Checking that the scores on the similation-2 dataset are still
        # at least as good as they were when I finished the contest
        self.assertTrue(np.linalg.norm(coefficients['intercept_true'] - coefficients['intercept_pred']) < 1.62)
        self.assertTrue(np.linalg.norm(coefficients['effect_0_true'] - coefficients['effect_0_pred']) < 0.72)


from pathlib import Path

import json
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import matplotlib.dates as mdates

from skmixed.lme.models import LinearLMESparseModel

data_path = Path("/Users/aksh/Storage/repos/IHME_data/covid-19/seir-pipeline-outputs/")
covariates_version = "2020_05_09.01.10"
regression_version = "vis_test_3"
betas_path = data_path / 'regression' / regression_version / "betas"
coefficients_path = data_path / 'regression' / regression_version / "coefficients" / "coefficients_0.csv"
reg_settings_path = data_path / 'regression' / regression_version / "settings.json"
covariates_path = data_path / 'covariate' / covariates_version / "cached_covariates.csv"
cov_metadata_path = data_path / 'covariate' / covariates_version / "metadata.yaml"
location_metadata_path = data_path / 'metadata-inputs' / 'location_metadata_664.csv'


def format_xaxis(ax, start_date, end_date, major_tick_interval_days=14, margins_days=5):
    months = mdates.DayLocator(interval=major_tick_interval_days)
    days = mdates.DayLocator()  # Every day
    months_fmt = mdates.DateFormatter('%m/%d')

    # format the ticks
    ax.xaxis.set_major_locator(months)
    ax.xaxis.set_major_formatter(months_fmt)
    ax.xaxis.set_minor_locator(days)

    # round to nearest years.
    datemin = np.datetime64(start_date, 'D') - np.timedelta64(margins_days, 'D')
    datemax = np.datetime64(end_date, 'D') + np.timedelta64(margins_days, 'D')
    ax.set_xlim(datemin, datemax)

    # format the coords message box
    ax.format_xdata = mdates.DateFormatter('%Y-%m-%d')

def launch_covid_experiment(num_groups=60):
    from matplotlib import rcParams

    rcParams['font.family'] = 'monospace'

    # Read the data
    num_groups = 15
    coefficients_df = pd.read_csv(coefficients_path)
    groups = list(coefficients_df['group_id'].unique())
    groups = groups[:num_groups]
    betas_df = {group: pd.read_csv(betas_path / f'{group}' / 'regression_draw_0.csv') for group in groups}
    reg_settings = json.load(open(reg_settings_path))
    covariates = list(reg_settings['covariates'].keys())
    covariates.remove("intercept")  # the skmixed model adds its own intercept
    location_metadata = pd.read_csv(location_metadata_path)
    id2loc = location_metadata.set_index('location_id')['location_name'].to_dict()

    # Combine all the data into one matrix
    all_betas = pd.concat(betas_df.values(), axis=0)
    all_betas["std"] = 0.1
    # all_betas[covariates] = normalize(all_betas[covariates].to_numpy(), axis=0)
    X = all_betas[["location_id"] + covariates + ["std"]].to_numpy()
    columns_labels = [0] + len(covariates) * [3] + [4]
    y = all_betas["beta"].to_numpy()
    print("done data")

    # Fit the model
    model = LinearLMESparseModel(lb=0, lg=0, initializer="EM", n_iter=100, n_iter_inner=100, n_iter_outer=1)
    model.fit(X, y, columns_labels=columns_labels)
    all_betas["prediction"] = model.predict(X, columns_labels=columns_labels)
    logger = model.logger_
    print("done fitting dense")

    model_sparse = LinearLMESparseModel(lb=0, lg=0, nnz_tbeta=3, nnz_tgamma=2,
                                        regularization_type="l2",
                                        initializer="EM", n_iter=100, n_iter_inner=100, n_iter_outer=15)
    model_sparse.fit(X, y, columns_labels=columns_labels)
    all_betas["sparse_prediction"] = model_sparse.predict(X, columns_labels=columns_labels,
                                                          use_sparse_coefficients=True)
    print("done fitting sparse")
    #
    model_smart = LinearLMESparseModel(lb=0, lg=0, nnz_tbeta=3, nnz_tgamma=2,
                                       regularization_type="loss-weighted",
                                       initializer="EM", n_iter=100, n_iter_inner=100, n_iter_outer=15)
    model_smart.fit(X, y, columns_labels=columns_labels)
    all_betas["weighted_sparse_prediction"] = model_smart.predict(X, columns_labels=columns_labels,
                                                                  use_sparse_coefficients=True)
    print("done fitting smart")

    # Evaluate the model
    def print_on_plot(elements, ax, x0=0.2, y0=9, h=0.4):
        offset = 0
        for s in elements:
            if type(s) is tuple:
                ax.text(x0, y0 - offset, s[0], color=s[1])
                offset += h
            elif type(s) is str:
                ax.text(x0, y0 - offset, s)
                offset += h
            else:
                raise ValueError("wrong type: %s" % type(s))

    fig = plt.figure(figsize=(12, 7 * len(groups)))
    grid = plt.GridSpec(len(groups), 2)
    counter = 0
    ihme_scores = []
    me_scores = []
    for i, group in enumerate(groups):
        ax = fig.add_subplot(grid[i, 0])
        cur_betas = all_betas[all_betas["location_id"] == group]
        time = pd.to_datetime(cur_betas['date'])
        ax.plot(time, cur_betas["beta"], label='True')
        ax.plot(time, cur_betas["beta_pred"], label='IHME')
        ax.plot(time, cur_betas["sparse_prediction"], label="R&S Sparse")
        ax.plot(time, cur_betas["prediction"], label="Dense MM")
        ax.plot(time, cur_betas["weighted_sparse_prediction"], label="R&S + W Sparse")
        ax.legend()
        start_date = time.to_list()[0]
        end_date = time.to_list()[-1]
        format_xaxis(ax, start_date, end_date)
        ax.set_title(f"{regression_version}: {id2loc[group]}")
        ax2 = fig.add_subplot(grid[i, 1])
        ax2.set_xlim((0, 12))
        ax2.set_ylim((0, 12))
        table_format = "%-21s%10.2e%10.2e%10.2e%10.2e"
        ihme_error = np.linalg.norm(cur_betas["beta"] - cur_betas["beta_pred"])
        dense_error = np.linalg.norm(cur_betas["beta"] - cur_betas["prediction"])
        sparse_error = np.linalg.norm(cur_betas["beta"] - cur_betas["sparse_prediction"])
        weighted_sparse_error = np.linalg.norm(cur_betas["beta"] - cur_betas["weighted_sparse_prediction"])

        ihme_scores.append(ihme_error)
        me_scores.append(dense_error)

        effect2coef = [0, 2, 3, 4, 5]

        def coef_to_color(beta, gamma):
            if beta == 0:
                return "red"
            elif gamma == 0:
                return "blue"
            else:
                return "black"

        sorted_errors_idx = np.argsort([ihme_error, dense_error, sparse_error, weighted_sparse_error])
        colors_for_errors = ["", "", "", ""]
        for k, color in enumerate(["green", "blue", "black", "black"]):
            colors_for_errors[sorted_errors_idx[k]] = color

        # if ihme_error > weighted_sparse_error:
        #    counter += 1
        statistics = [
                         "RMSE:",
                         # "%-12s: %.2e" % ("  IHME", ihme_error),
                         # "%-12s: %.2e" % ("  Dense MM", dense_error),
                         ("%-12s: %.2e" % ("  IHME", ihme_error), colors_for_errors[0]),
                         ("%-12s: %.2e" % ("  Dense MM", dense_error), colors_for_errors[1]),
                         ("%-12s: %.2e" % ("  R&S", sparse_error), colors_for_errors[2]),
                         ("%-12s: %.2e" % ("  R&S + W", weighted_sparse_error), colors_for_errors[3]),
                         "",
                     ] + \
                     [
                         "Full MM Coefficients:",
                         "%-21s%10s%10s%10s%10s" % ("name", "local", "mean", "RE", "Var"),
                     ] + \
                     [
                         table_format % (""
                                         "  " + covariate,
                                         model.coef_["per_group_coefficients"][i, effect2coef[j]],
                                         model.coef_["beta"][j],
                                         model.coef_["random_effects"][i, j],
                                         model.coef_["gamma"][j])
                         for j, covariate in enumerate(["intercept"] + covariates)
                     ] + \
                     [
                         "\n",
                         "Sparse Relax-and-Split Coefficients:",
                         "%-21s%10s%10s%10s%10s" % ("name", "local", "mean", "RE", "Var"),
                     ] + \
                     [
                         (table_format % (""
                                          "  " + covariate,
                                          model_sparse.coef_["sparse_per_group_coefficients"][i, effect2coef[j]],
                                          model_sparse.coef_["tbeta"][j],
                                          model_sparse.coef_["sparse_random_effects"][i, j],
                                          model_sparse.coef_["tgamma"][j]),
                          coef_to_color(model_sparse.coef_["tbeta"][j], model_sparse.coef_["tgamma"][j]))
                         for j, covariate in enumerate(["intercept"] + covariates)
                     ] + \
                     [
                         "\n",
                         "W-norm Sparse Relax-and-Split Coefficients:",
                         "%-21s%10s%10s%10s%10s" % ("name", "local", "mean", "RE", "Var"),
                     ] + \
                     [
                         (table_format % (""
                                          "  " + covariate,
                                          model_smart.coef_["sparse_per_group_coefficients"][i, effect2coef[j]],
                                          model_smart.coef_["tbeta"][j],
                                          model_smart.coef_["sparse_random_effects"][i, j],
                                          model_smart.coef_["tgamma"][j]),
                          coef_to_color(model_smart.coef_["tbeta"][j], model_smart.coef_["tgamma"][j]))
                         for j, covariate in enumerate(["intercept"] + covariates)
                     ]

        print_on_plot(statistics, ax2, x0=-1, y0=12)
        ax2.axis('off')

    plt.savefig("fit_0.png")

    print(counter / len(groups))
    plt.figure(figsize=(8, 8))
    plt.scatter(ihme_scores, me_scores)
    plt.xlabel("RMSE IHME")
    plt.ylabel("RMSE Aleksei")
    max_err = max(max(ihme_scores), max(me_scores))
    plt.plot([0, max_err], [0, max_err], '--b')
    plt.savefig("comparison_0.pdf")

if __name__ == "__main__":
    launch_covid_experiment()


# This code implements solvers for linear mixed-effects models.
# Copyright (C) 2020 Aleksei Sholokhov, aksh@uw.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from sklearn.base import clone, BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_consistent_length, check_is_fitted
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem


class Tree(BaseEstimator, RegressorMixin):
    def __init__(self, model: LinearLMESparseModel, max_depth=3, information_criterion="IC_vaida2005aic",
                 minimal_gain=1e1):
        self.model = model
        self.max_depth = max_depth
        self.information_criterion = information_criterion
        self.minimal_gain = minimal_gain

    def fit_problem(self, problem):
        chosen_features = [0]
        possible_features = np.arange(1, problem.num_categorical_features)
        problem = problem.pivot((0,))
        model = clone(self.model)
        model.logger_keys = model.logger_keys + (self.information_criterion,)
        x, y = problem.to_x_y()
        model.fit(x, y)  # TODO: fix it back to fit_problem

        previous_information_gain = model.logger_.get(self.information_criterion)
        for i in range(self.max_depth):
            all_gains = []
            all_solvers = []
            for j, k in enumerate(possible_features):           # Tree complexity stop
                current_problem = problem.pivot(set(chosen_features + [k]))
                if min(current_problem.groups_sizes) < 3:
                    all_gains.append(np.infty)
                    all_solvers.append(None)
                    continue
                current_solver = clone(model)
                current_solver.coef_ = model.coef_
                current_solver.fit_problem(current_problem, warm_start=True)
                all_gains.append(current_solver.logger_.get(self.information_criterion))
                all_solvers.append(current_solver)
            if all([gain == np.infty for gain in all_gains]):    # Granularity stop
                break
            best_gain_idx = np.argmin(all_gains)
            # assert len(best_gain_idx) == 1, "More than one best feature option"
            gain_difference = previous_information_gain - all_gains[best_gain_idx]
            if gain_difference <= self.minimal_gain:            # Information gain stop
                break
            chosen_feature = possible_features[best_gain_idx]
            chosen_features.append(chosen_feature)
            possible_features = possible_features[possible_features != chosen_feature]
            model = all_solvers[best_gain_idx]
            previous_information_gain = all_gains[best_gain_idx]

        self.coef_ = {"chosen_categorical_features": set(chosen_features)}
        self.fitted_model_ = model

    def predict(self, x):
        # check_is_fitted(self, 'coef_')    # TODO: figure out why this check does not work
        problem = LinearLMEProblem.from_x_y(x)
        return self.predict_problem(problem)

    def predict_problem(self, problem):
        # check_is_fitted(self, 'coef_')    # same
        pivoted_problem = problem.pivot(self.coef_["chosen_categorical_features"])
        return  self.fitted_model_.predict_problem(pivoted_problem)


class Forest(BaseEstimator, RegressorMixin):

    def __init__(self, continuous_model, num_trees=10, max_depth=3, information_criterion="IC_vaida2005aic",
                 minimal_gain=1e1):
        self.model = continuous_model
        self.num_trees = num_trees
        self.max_depth = max_depth
        self.information_criterion=information_criterion
        self.minimal_gain=minimal_gain

    def fit_problem(self, problem: LinearLMEProblem, seed=42):
        self.trees_ = []
        self.bootstrap_idxs_ = []
        assert problem.answers is not None, "Problem does not include the target variable"
        for i in range(self.num_trees):
            bootstrap_problem = problem.bootstrap(seed=i+seed)
            tree = Tree(model=clone(self.model),
                        max_depth=self.max_depth,
                        information_criterion=self.information_criterion,
                        minimal_gain=self.minimal_gain)
            tree.fit_problem(bootstrap_problem)
            self.trees_.append(tree)
            self.bootstrap_idxs_.append(bootstrap_problem.categorical_features_bootstrap_idx)

    def get_ensemble_predictions(self, problem: LinearLMEProblem):
        # check_is_fitted(self, "trees_", "bootstrap_idxs_")
        predictions = []
        for tree, bootstrap_idxs in zip(self.trees_, self.bootstrap_idxs_):
            bootstrap_problem = problem.bootstrap(categorical_features_idx=bootstrap_idxs, do_bootstrap_objects=False)
            answers = tree.predict_problem(bootstrap_problem)
            predictions.append(answers)
        return predictions

    def predict_problem(self, problem: LinearLMEProblem):
        predictions = self.get_ensemble_predictions(problem)
        return np.mean(predictions, axis=0)

    def get_prediction_uncertainty(self, problem, percentile=5):
        assert 0 < percentile < 50, "Percentile should be between 0 and 50"
        predictions = np.array(self.get_ensemble_predictions(problem))
        lower_percentile = np.percentile(predictions, axis=0, q=percentile)
        higher_percentile = np.percentile(predictions, axis=0, q=100-percentile)
        return lower_percentile, higher_percentile



from typing import Set


class Logger:
    """
    Helper class for logging the progress of iterative methods.
    """
    def __init__(self, list_of_keys: Set = ()):
        self.keys = list_of_keys
        self.dict = {key: [] for key in list_of_keys}

    def log(self, parameters):
        for key in self.keys:
            self.dict[key].append(parameters.get(key, None))
        return self

    def add(self, key, value):
        self.dict[key] = value
        return self

    def get(self, key):
        return self.dict[key]



import unittest

import numpy as np

from skmixed.lme.problems import LinearLMEProblem


class TestLinearLMEProblem(unittest.TestCase):
    def test_correctness(self):
        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [3, 3, 3],
            "random_intercept": True,
            "features_covariance_matrix": np.array([
                [1, 0, 0],
                [0, 1, 0.7],
                [0, 0.7, 1]
            ]),
            "obs_std": 0.1,
        }
        problem, true_parameters = LinearLMEProblem.generate(**problem_parameters,
                                                             seed=0)
        x1, y1 = problem.to_x_y()
        problem2 = LinearLMEProblem.from_x_y(x1, y1)
        for i, (x, y, z, l) in enumerate(problem2):
            assert np.allclose(y, x.dot(true_parameters['beta']) + z.dot(true_parameters['random_effects'][i][1]) + true_parameters['errors'][i])

    def test_creation_and_from_to_x_y(self):
        problem, true_parameters = LinearLMEProblem.generate(groups_sizes=[20, 30, 50],
                                                             features_labels=[3, 5, 3, 1, 2, 6],
                                                             random_intercept=True,
                                                             obs_std=0.1,
                                                             seed=42)
        x1, y1 = problem.to_x_y()

        problem2 = LinearLMEProblem.from_x_y(x1, y1)
        x2, y2 = problem2.to_x_y()
        self.assertTrue(np.all(x1 == x2) and np.all(y1 == y2))
        test_problem, true_test_parameters = LinearLMEProblem.generate(groups_sizes=[3, 4, 5],
                                                                       features_labels=[3, 5, 3, 1, 2, 6],
                                                                       random_intercept=True,
                                                                       beta=true_parameters["beta"],
                                                                       gamma=true_parameters["gamma"],
                                                                       true_random_effects=true_parameters[
                                                                           "random_effects"],
                                                                       obs_std=0.1,
                                                                       seed=43)

        self.assertTrue(np.all(true_parameters["beta"] == true_test_parameters["beta"])
                        and np.all(true_parameters["gamma"] == true_test_parameters["gamma"]))
        test_us = dict(true_test_parameters["random_effects"])
        for k, u1 in true_parameters["random_effects"]:
            u2 = test_us.get(k, None)
            if u2 is not None:
                self.assertTrue(np.all(u1 == u2))

    def test_creation_from_no_data(self):
        problem, true_parameters = LinearLMEProblem.generate(groups_sizes=[4, 5, 10],
                                                             features_labels=[],
                                                             random_intercept=True,
                                                             obs_std=0.1,
                                                             seed=42)

        self.assertEqual(len(true_parameters["beta"]), 1, "Beta should be of len = 1 for no-data problem")
        self.assertEqual(len(true_parameters["gamma"]), 1), "Gamma should be of len = 1 for no-data problem"
        self.assertTrue(np.all([np.all(x == 1) and np.all(z == 1) for x, y, z, l in
                                problem])), "All fixed and random features should be 1 for no-data problem"

    def test_from_to_xy_preserves_dataset_structure(self):
        study_sizes = [20, 15, 10]
        num_studies = sum(study_sizes)
        num_features = 6
        num_random_effects = 4
        num_categorical_effects = 2
        np.random.seed(42)
        x = np.random.rand(num_studies + 1, 1 + (num_features - 1) + 1 + (num_random_effects - 1) + 1)
        categorical_features = np.random.randint(0, 2, (num_studies+1, num_categorical_effects))
        x = np.concatenate((x, categorical_features), axis=1)
        y = np.random.rand(num_studies)
        x[1:, 0] = np.repeat([0, 1, 2], study_sizes)
        x[0, :] = [0] + [1] * (num_features - 1) + [3] + [2] * (num_random_effects - 1) + [4] + [5]*num_categorical_effects
        np.random.shuffle(x[1:, :])
        problem = LinearLMEProblem.from_x_y(x, y)
        x2, y2 = problem.to_x_y()
        self.assertTrue(np.all(x2 == x), msg="x is not the same after from/to transformation")
        self.assertTrue(np.all(y2 == y), msg="y is not the same after from/to transformation")

    def test_pivoting(self):
        problem, true_parameters = LinearLMEProblem.generate(groups_sizes=[40, 30, 50],
                                                             features_labels=[3, 3, 1, 2, 6, 5],
                                                             random_intercept=True,
                                                             obs_std=0.1,
                                                             seed=42)
        x1, y1 = problem.to_x_y()
        problem2 = problem.pivot((0, 1))
        unique_pivot_groups = set([tuple(s) for s in x1[1:, np.array([1, 6])]])
        assert problem2.num_groups == len(unique_pivot_groups)
        x2, y2 = problem2.to_x_y()
        problem3 = LinearLMEProblem.from_x_y(x2, y2).pivot((0, ))
        assert problem3.num_groups == 3
        x3, y3 = problem3.to_x_y()
        self.assertTrue(np.all(x3 == x1), msg="x is not the same after from/to transformation")
        self.assertTrue(np.all(y3 == y1), msg="y is not the same after from/to transformation")


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem
from skmixed.lme.trees import Tree


class TestTree(unittest.TestCase):

    def test_selecting_categorical_features_only(self):
        # TODO: make a better test for this functionality
        assert True
        return None
        max_mse = 0.05
        min_explained_variance = 0.9
        categorical_effects_min_accuracy = 0.7
        active_categorical_features = (0, 2, 3)
        for i in range(1):
            problem, true_parameters = LinearLMEProblem.generate(groups_sizes=[40, 30, 50],
                                                                 features_labels=[3, 6, 5, 5, 6],
                                                                 random_intercept=True,
                                                                 obs_std=0.1,
                                                                 seed=i)
            continuous_model = LinearLMESparseModel(lb=0, lg=0, nnz_tbeta=2, nnz_tgamma=2)
            tree_model = Tree(model=continuous_model, max_depth=3)
            tree_model.fit_problem(problem)
            y_pred = tree_model.predict_problem(problem)
            pivoted_problem = problem.pivot(tree_model.coef_['chosen_categorical_features'])
            correctly_pivoted_problem = problem.pivot(active_categorical_features)
            x_true, y_true = problem.to_x_y()
            explained_variance = explained_variance_score(y_true, y_pred)
            mse = mean_squared_error(y_true, y_pred)
            selection_accuracy = np.mean([feature in active_categorical_features for feature in tree_model.coef_['chosen_categorical_features']])
            # for x, y, z, l in problem:
            #     plt.scatter(x[:, 1], y)
            # plt.title("Original problem")
            # plt.show()
            # for x, y, z, l in correctly_pivoted_problem:
            #     plt.scatter(x[:, 1], y)
            # plt.title("Correctly pivoted problem")
            # plt.show()


            self.assertGreater(explained_variance, min_explained_variance,
                               msg="%d) Explained variance is too small: %.3f < %.3f. (seed=%d)"
                                   % (i,
                                      explained_variance,
                                      min_explained_variance,
                                      i))
            self.assertGreater(max_mse, mse,
                               msg="%d) MSE is too big: %.3f > %.2f  (seed=%d)"
                                   % (i,
                                      mse,
                                      max_mse,
                                      i))
            self.assertGreater(selection_accuracy, categorical_effects_min_accuracy,
                               msg="%d) Fixed Effects Selection Accuracy is too small: %.3f < %.2f  (seed=%d)"
                                   % (i,
                                      selection_accuracy,
                                      categorical_effects_min_accuracy,
                                      i)
                               )


# This code implements solvers for linear mixed-effects models.
# Copyright (C) 2020 Aleksei Sholokhov, aksh@uw.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Set

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin

from skmixed.logger import Logger
from skmixed.lme.oracles import LinearLMEOracle

# TODO: implement sigma inference


class LMECascadeOracle:
    def __init__(self, levels, hierarchy):
        self.levels = levels
        self.hierarchy = hierarchy
        self.oracles = []
        for level in levels:
            self.oracles.append(LinearLMEOracle(level))

    def loss(self, beta, gamma):
        loss = 0
        for level, oracle in enumerate(self.oracles):
            loss += oracle.loss(beta, (level+1)*gamma)
        return loss

    def gradient_gamma(self, beta, gamma):
        gradient = np.zeros(len(gamma))
        for level, oracle in enumerate(self.oracles):
            gradient += (level+1)*oracle.gradient_gamma(beta, (level+1)*gamma)
        return gradient

    def optimal_beta(self, gamma):
        total_kernel = 0
        total_tail = 0
        for level, oracle in enumerate(self.oracles):
            kernel, tail = oracle.optimal_beta((level+1)*gamma, _dont_solve_wrt_beta=True)
            total_kernel += kernel
            total_tail += tail
        return np.linalg.solve(total_kernel, total_tail)

    def optimal_random_effects(self, beta, gamma):
        random_effects = {1: np.zeros(len(gamma))}
        for level, oracle in enumerate(self.oracles):
            for i, group in enumerate(oracle.problem.group_labels):
                parent = self.hierarchy[group]
                u = random_effects[parent]
                beta_current = beta.copy()
                for k in range(len(beta)):
                    j = int(oracle.beta_to_gamma_map[k])
                    if j >= 0:
                        beta_current[k] += u[j]
                random_effects[group] = oracle.optimal_random_effects(beta_current, gamma)[i] + u
        return random_effects


class LMECascade(BaseEstimator, RegressorMixin):
    def __init__(self,
                 tol: float = 1e-4,
                 tol_inner: float = 1e-4,
                 solver: str = "pgd",
                 n_iter: int = 1000,
                 n_iter_inner: int = 20,
                 use_line_search: bool = True,
                 logger_keys: Set = ('converged', 'loss',)):
        self.tol = tol
        self.tol_inner = tol_inner
        self.solver = solver
        self.n_iter = n_iter
        self.n_iter_inner = n_iter_inner
        self.use_line_search = use_line_search
        self.logger_keys = logger_keys

    def fit_problem(self, levels, hierarchy):
        oracle = LMECascadeOracle(levels, hierarchy)

        num_levels = len(levels)
        num_fixed_effects = levels[0].num_fixed_effects
        num_random_effects = levels[0].num_random_effects

        beta = np.ones(num_fixed_effects)
        gamma = 2*np.ones(num_random_effects)

        def projected_direction(current_gamma: np.ndarray, current_direction: np.ndarray):
            proj_direction = current_direction.copy()
            for j, _ in enumerate(current_gamma):
                if current_gamma[j] <= 1e-15 and current_direction[j] <= 0:
                    proj_direction[j] = 0
            return proj_direction

        loss = oracle.loss(beta, gamma)
        self.logger_ = Logger(self.logger_keys)

        prev_beta = np.infty
        prev_gamma = np.infty
        iteration = 0

        while (np.linalg.norm(beta - prev_beta) > self.tol
                    or np.linalg.norm(gamma - prev_gamma) > self.tol) and iteration < self.n_iter:

                if iteration >= self.n_iter:
                    us = oracle.optimal_random_effects(beta, gamma)
                    if len(self.logger_keys) > 0:
                        self.logger_.log(**locals())
                    self.coef_ = {"beta": beta,
                                  "gamma": gamma,
                                  "random_effects": us
                                  }
                    self.logger_.add("converged", 0)
                    return self

                if self.solver == 'pgd':

                    prev_beta = beta
                    prev_gamma = gamma

                    beta = oracle.optimal_beta(gamma)

                    # =============== INNER (GAMMA) ITERATION ===========
                    gradient_gamma = oracle.gradient_gamma(beta, gamma)
                    # projecting the gradient to the set of constraints
                    direction = projected_direction(gamma, -gradient_gamma)

                    inner_iteration = 0
                    while (np.linalg.norm(direction) > self.tol_inner
                           and inner_iteration < self.n_iter_inner):
                        if self.use_line_search:
                            # line search method
                            step_len = 0.1
                            for i, _ in enumerate(gamma):
                                if direction[i] < 0:
                                    step_len = min(-gamma[i] / direction[i], step_len)

                            current_loss = oracle.loss(beta, gamma)

                            while (oracle.loss(beta, gamma + step_len * direction)
                                   >= (1 - np.sign(current_loss) * 1e-5) * current_loss):
                                step_len *= 0.5
                                if step_len <= 1e-15:
                                    break
                        else:
                            # fixed step size
                            step_len = 1 / iteration
                        if step_len <= 1e-15:
                            break
                        gamma = gamma + step_len * direction
                        gradient_gamma = oracle.gradient_gamma(beta, gamma)
                        direction = projected_direction(gamma, -gradient_gamma)
                        inner_iteration += 1

                iteration += 1
                loss = oracle.loss(beta, gamma)
                if len(self.logger_keys) > 0:
                    self.logger_.log(locals())

        us = oracle.optimal_random_effects(beta, gamma)

        self.logger_.add('converged', 1)
        # self.logger_.add('iterations', iteration)

        self.coef_ = {
            "beta": beta,
            "gamma": gamma,
            "random_effects": us,
        }

        return self



# This code implements solvers for linear mixed-effects models.
# Copyright (C) 2020 Aleksei Sholokhov, aksh@uw.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Set

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_consistent_length, check_is_fitted

from skmixed.lme.problems import LinearLMEProblem
from skmixed.lme.oracles import LinearLMEOracleRegularized, LinearLMEOracleW
from skmixed.logger import Logger
from skmixed.helpers import get_per_group_coefficients


class LinearLMESparseModel(BaseEstimator, RegressorMixin):
    """
    Solve regularized sparse Linear Mixed Effects problem with projected gradient descent method.

    The log-likelihood minimization problem which this model solves is::

        min ℋ(β, 𝛄, tβ, t𝛄) w.r.t. all four arguments (β, 𝛄, tβ, t𝛄)
        s.t. nnz(tbeta) <= nnz_tbeta and nnz(gamma) <= nnz_tgamma where

        ℋ(β, 𝛄, tβ, t𝛄) := ℒ(β, 𝛄) + lb/2*||β - tβ||^2 + lg/2*||𝛄 - t𝛄||^2

        ℒ(β, 𝛄) = ∑(yi - Xi*β)ᵀΩi^{-1}(yi - Xi*β) + ln(det(Ωi))

        Ωi = Zi*diag(𝛄)Ziᵀ + diag(obs_stds)

    The original statistical model which this loss is based on is::

        Y_i = X_i*β + Z_i*u_i + 𝜺_i,

        where

        u_i ~ 𝒩(0, diag(𝛄)),

        𝛄 ~ 𝒩(t𝛄, 1/lg)

        β ~ 𝒩(tβ, 1/lb)

        𝜺_i ~ 𝒩(0, diag(obs_std)

    See my paper for more details.
    """

    def __init__(self,
                 tol: float = 1e-4,
                 tol_inner: float = 1e-4,
                 tol_outer: float = 1e-2,
                 solver: str = "pgd",
                 initializer=None,
                 n_iter: int = 1000,
                 n_iter_inner: int = 20,
                 n_iter_outer: int = 1,
                 use_line_search: bool = True,
                 lb: float = 1,
                 lg: float = 1,
                 regularization_type: str = "l2",
                 nnz_tbeta: int = 3,
                 nnz_tgamma: int = 3,
                 logger_keys: Set = ('converged',)):
        """
        init: initializes the model.

        Parameters
        ----------
        tol : float
            Tolerance for stopping criterion: ||tβ_{k+1} - tβ_k|| <= tol and ||t𝛄_{k+1} - t𝛄_k|| <= tol.

        tol_inner : float
            Tolerance for inner optimization subroutine (min ℋ w.r.t. 𝛄) stopping criterion:
            ||projected ∇ℋ|| <= tol_inner

        solver : {'pgd'} Solver to use in computational routines:

                - 'pgd' : Projected Gradient Descent

        initializer : {None, 'EM'}, Optional
            Whether to use an initializer before starting the main optimization routine:

                - None : Does not do any special initialization, starts with the given initial point.
                - 'EM' : Performs one step of a naive EM-algorithm in order to improve the initial point.

        n_iter : int
            Number of iterations for the outer optimization cycle.

        n_iter_inner : int
            Number of iterations for the inner optimization cycle.

        use_line_search : bool, default = True
            Whether to use line search when optimizing w.r.t. 𝛄. If true, it starts from step_len = 1 and cuts it
            in half until the descent criterion is met. If false, it uses a fixed step size of 1/iteration_number.

        lb : float
            Regularization coefficient for the tβ-related term, see the loss-function description.

        lg : float
            Regularization coefficient for the t𝛄-related term, see the loss-function description.

        regularization_type : str, one of {"l2", "loss-weighted"}, default = "l2"
            Type of norm in the regularization terms. Options are:

                - 'l2' : Euclidean norm.
                - 'loss-weighted' : A weighted norm designed so to drop the least important coefficients.

        nnz_tbeta : int,
            How many non-zero coefficients are allowed in tβ.

        nnz_tgamma : int,
            How many non-zero coefficients are allowed in t𝛄.
        """

        self.tol = tol
        self.tol_inner = tol_inner
        self.tol_outer = tol_outer
        self.solver = solver
        self.initializer = initializer
        self.n_iter = n_iter
        self.n_iter_inner = n_iter_inner
        self.n_iter_outer = n_iter_outer
        self.use_line_search = use_line_search
        self.lb = lb
        self.lg = lg
        self.nnz_tbeta = nnz_tbeta
        self.nnz_tgamma = nnz_tgamma
        self.logger_keys = logger_keys
        self.regularization_type = regularization_type

    def fit_problem(self,
                    problem: LinearLMEProblem,
                    columns_labels: np.ndarray = None,
                    initial_parameters: dict = None,
                    warm_start=False,
                    random_intercept=True,
                    **kwargs):
        """
        Fits a Linear Model with Linear Mixed-Effects to the given data.

        Parameters
        ----------
        problem : LinearLMEProblem
            The problem to fit the model into

        columns_labels : np.ndarray
            List of column labels. There shall be only one column of group labels and answers STDs,
            and overall n columns with fixed effects (1 or 3) and k columns of random effects (2 or 3).

                - 1 : fixed effect
                - 2 : random effect
                - 3 : both fixed and random,
                - 0 : groups labels
                - 4 : answers standard deviations

        initial_parameters : np.ndarray
            Dict with possible fields:

                -   | 'beta0' : np.ndarray, shape = [n],
                    | Initial estimate of fixed effects. If None then it defaults to an all-ones vector.
                -   | 'gamma0' : np.ndarray, shape = [k],
                    | Initial estimate of random effects covariances. If None then it defaults to an all-ones vector.
                -   | 'tbeta0' : np.ndarray, shape = [n],
                    | Initial estimate of sparse fixed effects. If None then it defaults to an all-zeros vector.
                -   | 'tgamma0' : np.ndarray, shape = [k],
                    | Initial estimate of sparse random covariances. If None then it defaults to an all-zeros vector.

        warm_start : bool, default is False
            Whether to use previous parameters as initial ones. Overrides initial_parameters if given.
            Throws NotFittedError if set to True when not fitted.

        random_intercept : bool, default = True
            Whether treat the intercept as a random effect.
        kwargs :
            Not used currently, left here for passing debugging parameters.

        Returns
        -------
        self : LinearLMESparseModel
            Fitted regression model.
        """

        if initial_parameters is None:
            initial_parameters = {}
        beta0 = initial_parameters.get("beta", None)
        gamma0 = initial_parameters.get("gamma", None)
        tbeta0 = initial_parameters.get("tbeta", None)
        tgamma0 = initial_parameters.get("tgamma", None)
        _check_input_consistency(problem, beta0, gamma0, tbeta0, tgamma0)

        if self.regularization_type == "l2":
            oracle = LinearLMEOracleRegularized(problem,
                                                lb=self.lb,
                                                lg=self.lg,
                                                nnz_tbeta=self.nnz_tbeta,
                                                nnz_tgamma=self.nnz_tgamma
                                                )
        elif self.regularization_type == "loss-weighted":
            oracle = LinearLMEOracleW(problem,
                                      lb=self.lb,
                                      lg=self.lg,
                                      nnz_tbeta=self.nnz_tbeta,
                                      nnz_tgamma=self.nnz_tgamma
                                      )
        else:
            raise ValueError("regularization_type is not understood.")

        num_fixed_effects = problem.num_fixed_effects
        num_random_effects = problem.num_random_effects
        assert num_fixed_effects >= self.nnz_tbeta
        assert num_random_effects >= self.nnz_tgamma
        # old_oracle = OldOracle(problem, lb=self.lb, lg=self.lg, k=self.nnz_tbeta, j=self.nnz_tgamma)

        if warm_start:
            check_is_fitted(self, 'coef_')
            beta = self.coef_["beta"]
            gamma = self.coef_["gamma"]
            tbeta = self.coef_["tbeta"]
            tgamma = self.coef_["tgamma"]

        else:
            if beta0 is not None:
                beta = beta0
            else:
                beta = np.ones(num_fixed_effects)

            if gamma0 is not None:
                gamma = gamma0
            else:
                gamma = np.ones(num_random_effects)

            if tbeta0 is not None:
                tbeta = tbeta0
            else:
                tbeta = np.zeros(num_fixed_effects)

            if tgamma0 is not None:
                tgamma = tgamma0
            else:
                tgamma = np.zeros(num_random_effects)

        if self.initializer == "EM":
            beta = oracle.optimal_beta(gamma, tbeta, beta=beta)
            us = oracle.optimal_random_effects(beta, gamma)
            gamma = np.sum(us ** 2, axis=0) / oracle.problem.num_groups
            # tbeta = oracle.optimal_tbeta(beta)
            # tgamma = oracle.optimal_tgamma(tbeta, gamma)

        def projected_direction(current_gamma, current_direction):
            proj_direction = current_direction.copy()
            for j, _ in enumerate(current_gamma):
                if current_gamma[j] <= 1e-15 and current_direction[j] <= 0:
                    proj_direction[j] = 0
            return proj_direction

        loss = oracle.loss(beta, gamma, tbeta, tgamma)
        self.logger_ = Logger(self.logger_keys)

        outer_iteration = 0

        while (outer_iteration < self.n_iter_outer
               and (np.linalg.norm(beta - tbeta) > self.tol_outer
                    or np.linalg.norm(gamma - tgamma) > self.tol_outer)):

            prev_tbeta = np.infty
            prev_tgamma = np.infty
            prev_beta = np.infty
            prev_gamma = np.infty

            iteration = 0
            while ((np.linalg.norm(tbeta - prev_tbeta) > self.tol
                    or np.linalg.norm(tgamma - prev_tgamma) > self.tol
                    or np.linalg.norm(beta - prev_beta) > self.tol
                    or np.linalg.norm(gamma - prev_gamma) > self.tol)
                   and iteration < self.n_iter):

                if iteration >= self.n_iter:
                    us = oracle.optimal_random_effects(beta, gamma)
                    if len(self.logger_keys) > 0:
                        self.logger_.log(**locals())
                    self.coef_ = {"beta": beta,
                                  "gamma": gamma,
                                  "tbeta": tbeta,
                                  "tgamma": tgamma,
                                  "random_effects": us
                                  }
                    self.logger_.add("converged", 0)
                    return self

                if self.solver == 'pgd':
                    inner_iteration = 0

                    prev_beta = beta
                    prev_gamma = gamma
                    prev_tbeta = tbeta
                    prev_tgamma = tgamma

                    beta = oracle.optimal_beta(gamma, tbeta, beta=beta)
                    gradient_gamma = oracle.gradient_gamma(beta, gamma, tgamma)
                    direction = projected_direction(gamma, -gradient_gamma)
                    while (np.linalg.norm(direction) > self.tol_inner
                           and inner_iteration < self.n_iter_inner):
                        # gradient_gamma = oracle.gradient_gamma(beta, gamma, tgamma)
                        # projecting the gradient to the set of constraints
                        # direction = projected_direction(gamma, -gradient_gamma)
                        if self.use_line_search:
                            # line search method
                            step_len = 0.1
                            for i, _ in enumerate(gamma):
                                if direction[i] < 0:
                                    step_len = min(-gamma[i] / direction[i], step_len)

                            current_loss = oracle.loss(beta, gamma, tbeta, tgamma)

                            while (oracle.loss(beta, gamma + step_len * direction, tbeta, tgamma)
                                   >= (1 - np.sign(current_loss) * 1e-5) * current_loss):
                                step_len *= 0.5
                                if step_len <= 1e-15:
                                    break
                        else:
                            # fixed step size
                            step_len = 1 / iteration
                        if step_len <= 1e-15:
                            break
                        gamma = gamma + step_len * direction
                        gradient_gamma = oracle.gradient_gamma(beta, gamma, tgamma)
                        direction = projected_direction(gamma, -gradient_gamma)
                        inner_iteration += 1

                    tbeta = oracle.optimal_tbeta(beta=beta, gamma=gamma)
                    tgamma = oracle.optimal_tgamma(tbeta, gamma, beta=beta)
                    iteration += 1

                loss = oracle.loss(beta, gamma, tbeta, tgamma)
                if len(self.logger_keys) > 0:
                    self.logger_.log(locals())
            # gradient_at_tgamma = oracle.gradient_gamma(tbeta, tgamma, tgamma)
            # direction_at_sparse_point = projected_direction(tgamma, -gradient_at_tgamma)
            outer_iteration += 1
            oracle.lb = 2 * (1 + oracle.lb)
            oracle.lg = 2 * (1 + oracle.lg)

        us = oracle.optimal_random_effects(beta, gamma)
        sparse_us = oracle.optimal_random_effects(tbeta, tgamma)

        per_group_coefficients = get_per_group_coefficients(beta, us, labels=problem.column_labels)
        sparse_per_group_coefficients = get_per_group_coefficients(tbeta, sparse_us, labels=problem.column_labels)

        self.logger_.add('converged', 1)
        # self.logger_.add('iterations', iteration) # TODO: fix inner + outer iterations counter
        for key in self.logger_keys:
            if key.startswith("IC_"):
                self.logger_.add(key, oracle.get_ic(key, beta=beta, gamma=gamma, tbeta=tbeta, tgamma=tgamma))

        self.coef_ = {
            "beta": beta,
            "gamma": gamma,
            "tbeta": tbeta,
            "tgamma": tgamma,
            "random_effects": us,
            "sparse_random_effects": sparse_us,
            "group_labels": np.copy(problem.group_labels),
            "per_group_coefficients": per_group_coefficients,
            "sparse_per_group_coefficients": sparse_per_group_coefficients,
        }

        return self

    def predict(self, x, columns_labels=None, use_sparse_coefficients=False):
        """
        Makes a prediction if .fit(X, y) was called before and throws an error otherwise.

        Parameters
        ----------
        x : np.ndarray
            Data matrix. Should have the same format as the data which was used for fitting the model:
            the number of columns and the columns' labels should be the same. It may contain new groups, in which case
            the prediction will be formed using the fixed effects only.

        use_sparse_coefficients : bool, default is False
            If true then uses sparse coefficients, tbeta and tgamma, for making a prediction, otherwise uses
            beta and gamma.

        Returns
        -------
        y : np.ndarray
            Models predictions.
        """
        check_is_fitted(self, 'coef_')
        problem = LinearLMEProblem.from_x_y(x, y=None, columns_labels=columns_labels)
        return self.predict_problem(problem, use_sparse_coefficients=True)

    def predict_problem(self, problem, use_sparse_coefficients=False):
        """
        Makes a prediction if .fit(X, y) was called before and throws an error otherwise.

        Parameters
        ----------
        x : np.ndarray
            Data matrix. Should have the same format as the data which was used for fitting the model:
            the number of columns and the columns' labels should be the same. It may contain new groups, in which case
            the prediction will be formed using the fixed effects only.

        use_sparse_coefficients : bool, default is False
            If true then uses sparse coefficients, tbeta and tgamma, for making a prediction, otherwise uses
            beta and gamma.

        Returns
        -------
        y : np.ndarray
            Models predictions.
        """
        check_is_fitted(self, 'coef_')

        if use_sparse_coefficients:
            beta = self.coef_['tbeta']
            us = self.coef_['sparse_random_effects']
        else:
            beta = self.coef_['beta']
            us = self.coef_['random_effects']

        assert problem.num_fixed_effects == beta.shape[0], \
            "Number of fixed effects is not the same to what it was in the train data."

        assert problem.num_random_effects == us[0].shape[0], \
            "Number of random effects is not the same to what it was in the train data."

        group_labels = self.coef_['group_labels']
        answers = []
        for i, (x, _, z, stds) in enumerate(problem):
            label = problem.group_labels[i]
            idx_of_this_label_in_train = np.where(group_labels == label)
            assert len(idx_of_this_label_in_train) <= 1, "Group labels of the classifier contain duplicates."
            if len(idx_of_this_label_in_train) == 1:
                idx_of_this_label_in_train = idx_of_this_label_in_train[0]
                y = x.dot(beta) + z.dot(us[idx_of_this_label_in_train][0])
            else:
                # If we have not seen this group (so we don't have inferred random effects for this)
                # then we make a prediction with "expected" (e.g. zero) random effects
                y = x.dot(beta)
            answers.append(y)
        return np.concatenate(answers)

    def score(self, x, y, sample_weight=None):
        """
        Returns the coefficient of determination R^2 of the prediction.

        The coefficient R^2 is defined as (1 - u/v), where u is the residual sum
        of squares ((y_true - y_pred) ** 2).sum() and v is the
        total sum of squares ((y_true - y_true.mean()) ** 2).sum().
        The best possible score is 1.0 and it can be negative (because the model can be arbitrarily worse).
        A constant model that always predicts the expected value of y,
        disregarding the input features, would get a R^2 score of 0.0.

        Parameters
        ----------
        x : np.ndarray
            Data matrix. Should have the same format as the data which was used for fitting the model:
            the number of columns and the columns' labels should be the same. It may contain new groups, in which case
            the prediction will be formed using the fixed effects only.

        y : np.ndarray
            Answers, real-valued array.

        sample_weight : array_like, Optional
            Weights of samples for calculating the R^2 statistics.

        Returns
        -------
        r2_score : float
            R^2 score

        """

        y_pred = self.predict(x)
        u = ((y - y_pred) ** 2).sum()
        v = ((y - y.mean()) ** 2).sum()
        return 1 - u / v

    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            columns_labels: np.ndarray = None,
            initial_parameters: dict = None,
            warm_start=False,
            random_intercept=True,
            **kwargs):
        """
        Fits a Linear Model with Linear Mixed-Effects to the given data.

        Parameters
        ----------
        x : np.ndarray
            Data. If columns_labels = None then it's assumed that columns_labels are in the first row of x.

        y : np.ndarray
            Answers, real-valued array.

        columns_labels : np.ndarray
            List of column labels. There shall be only one column of group labels and answers STDs,
            and overall n columns with fixed effects (1 or 3) and k columns of random effects (2 or 3).

                - 1 : fixed effect
                - 2 : random effect
                - 3 : both fixed and random,
                - 0 : groups labels
                - 4 : answers standard deviations

        initial_parameters : np.ndarray
            Dict with possible fields:

                -   | 'beta0' : np.ndarray, shape = [n],
                    | Initial estimate of fixed effects. If None then it defaults to an all-ones vector.
                -   | 'gamma0' : np.ndarray, shape = [k],
                    | Initial estimate of random effects covariances. If None then it defaults to an all-ones vector.
                -   | 'tbeta0' : np.ndarray, shape = [n],
                    | Initial estimate of sparse fixed effects. If None then it defaults to an all-zeros vector.
                -   | 'tgamma0' : np.ndarray, shape = [k],
                    | Initial estimate of sparse random covariances. If None then it defaults to an all-zeros vector.

        warm_start : bool, default is False
            Whether to use previous parameters as initial ones. Overrides initial_parameters if given.
            Throws NotFittedError if set to True when not fitted.

        random_intercept : bool, default = True
            Whether treat the intercept as a random effect.
        kwargs :
            Not used currently, left here for passing debugging parameters.

        Returns
        -------
        self : LinearLMESparseModel
            Fitted regression model.
        """
        problem = LinearLMEProblem.from_x_y(x, y, columns_labels, random_intercept=random_intercept, **kwargs)
        return self.fit_problem(problem,
                                columns_labels=columns_labels,
                                initial_parameters=initial_parameters,
                                warm_start=warm_start,
                                random_intercept=random_intercept)


def _check_input_consistency(problem, beta=None, gamma=None, tbeta=None, tgamma=None):
    """
    Checks the consistency of .fit() arguments

    Parameters
    ----------
    problem : LinearLMEProblem
        The problem which contains data
    beta : array-like, shape = [n], Optional
        Vector of fixed effects
    gamma : array-like, shape = [k], Optional
        Vector of random effects
    tbeta : array-like, shape = [n], Optional
        Vector of the sparse set of fixed effects (for regularized models)
    tgamma : array-like, shape = [k], Optional
        Vector of the sparse set of random effects (for regularized models)

    Returns
    -------
        output : None
            None if all the checks are passed, otherwise raises an exception
    """

    num_features = problem.num_fixed_effects
    num_random_effects = problem.num_random_effects
    if beta is not None:
        if tbeta is not None:
            check_consistent_length(beta, tbeta)
        assert len(beta) == num_features, "len(beta) is %d, but the number of features in data is %d" % (
            len(beta), num_features
        )
    if gamma is not None:
        if tgamma is not None:
            check_consistent_length(gamma, tgamma)
        assert len(gamma) == num_random_effects, "len(gamma) is %d, but the number of random effects in data is %d" % (
            len(gamma), num_random_effects
        )
    return None



# This code implements linear mixed-effects oracle as a subroutine for skmixed's subroutines.
# Copyright (C) 2020 Aleksei Sholokhov, aksh@uw.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Callable

import numpy as np
import scipy as sp
from scipy.linalg.lapack import get_lapack_funcs

from skmixed.lme.problems import LinearLMEProblem


class LinearLMEOracle:
    """
    Implements Linear Mixed-Effects Model functional for given problem.

    The model is::

        Y_i = X_i*β + Z_i*u_i + 𝜺_i,

        where

        u_i ~ 𝒩(0, diag(𝛄)),

        𝜺_i ~ 𝒩(0, Λ)

    The problem should be provided as LinearLMEProblem.

    """

    def __init__(self, problem: LinearLMEProblem):
        """
        Creates an oracle on top of the given problem

        Parameters
        ----------
        problem : LinearLMEProblem
            set of data and answers. See docs for LinearLMEProblem class for more details.
        """

        self.problem = problem
        self.omega_cholesky_inv = []
        self.omega_cholesky = []
        self.gamma = None
        beta_to_gamma_map = np.zeros(self.problem.num_fixed_effects)
        beta_counter = 0
        gamma_counter = 0
        for label in self.problem.column_labels:
            if label == 1:
                beta_to_gamma_map[beta_counter] = -1
                beta_counter += 1
            elif label == 2:
                gamma_counter += 1
            elif label == 3:
                beta_to_gamma_map[beta_counter] = gamma_counter
                beta_counter += 1
                gamma_counter += 1
            else:
                continue
        self.beta_to_gamma_map = beta_to_gamma_map

    def _recalculate_cholesky(self, gamma: np.ndarray):
        """
        Recalculates Cholesky factors of all Ω_i's when gamma changes.

        Supplementary subroutine which recalculates Cholesky decompositions and their inverses of matrices Ω_i
        when 𝛄 (estimated set of covariances for random effects) changes:

        Ω_i = Z_i*diag(𝛄)*Z_i^T + Λ_i = L_i*L_i^T

        Parameters
        ----------
        gamma : np.ndarray, shape=[k]
            vector of covariances for random effects

        Returns
        -------
            None :
                if all the Cholesky factors were updated and stored successfully, otherwise raises and error
        """

        if (self.gamma != gamma).any():
            self.omega_cholesky = []
            self.omega_cholesky_inv = []
            gamma_mat = np.diag(gamma)
            invert_upper_triangular: Callable[[np.ndarray], np.ndarray] = get_lapack_funcs("trtri")
            for x, y, z, stds in self.problem:
                omega = z.dot(gamma_mat).dot(z.T) + np.diag(stds)
                L = np.linalg.cholesky(omega)
                L_inv = invert_upper_triangular(L.T)[0].T
                self.omega_cholesky.append(L)
                self.omega_cholesky_inv.append(L_inv)
            self.gamma = gamma
        return None

    def loss(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> float:
        """
        Returns the loss function value ℒ(β, 𝛄).

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
            result : float
                The value of the loss function: ℒ(β, 𝛄)
        """

        result = 0
        self._recalculate_cholesky(gamma)
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            xi = y - x.dot(beta)
            result += 1 / 2 * np.sum(L_inv.dot(xi) ** 2) - np.sum(np.log(np.diag(L_inv)))
        return result

    def demarginalized_loss(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> float:
        result = 0
        self._recalculate_cholesky(gamma)
        us = self.optimal_random_effects(beta, gamma)
        for (x, y, z, stds), u in zip(self.problem, us):
            r = y - x.dot(beta) - z.dot(u)
            result += 1 / 2 * sum(r ** 2 / stds) + 1 / 2 * sum(np.log(stds))
        return result

    def gradient_gamma(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> np.ndarray:
        """
        Returns the gradient of the loss function with respect to gamma: ∇_𝛄[ℒ](β, 𝛄)

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
            grad_gamma: np.ndarray, shape = [k]
                The gradient of the loss function with respect to gamma: ∇_𝛄[ℒ](β, 𝛄)
        """

        self._recalculate_cholesky(gamma)
        grad_gamma = np.zeros(len(gamma))
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            xi = y - x.dot(beta)
            Lz = L_inv.dot(z)
            grad_gamma += 1 / 2 * np.sum(Lz ** 2, axis=0) - 1 / 2 * Lz.T.dot(L_inv.dot(xi)) ** 2
        return grad_gamma

    def hessian_gamma(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> np.ndarray:
        """
        Returns the Hessian of the loss function with respect to gamma ∇²_𝛄[ℒ](β, 𝛄).

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
            hessian: np.ndarray, shape = [k, k]
                Hessian of the loss function with respect to gamma ∇²_𝛄[ℒ](β, 𝛄).
        """
        self._recalculate_cholesky(gamma)
        num_random_effects = self.problem.num_random_effects
        hessian = np.zeros(shape=(num_random_effects, num_random_effects))
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            xi = y - x.dot(beta)
            Lz = L_inv.dot(z)
            Lxi = L_inv.dot(xi).reshape((len(xi), 1))
            hessian += (-Lz.T.dot(Lz) + 2 * (Lz.T.dot(Lxi).dot(Lxi.T).dot(Lz))) * (Lz.T.dot(Lz))
        return 1 / 2 * hessian

    def gradient_beta(self, beta: np.ndarray, gamma: np.ndarray, **kwargs):
        self._recalculate_cholesky(gamma)
        gradient = 0
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            xi = y - x.dot(beta)
            gradient += - (L_inv.dot(x)).T.dot(L_inv.dot(xi))
        return gradient

    def x_to_beta_gamma(self, x):
        beta = x[:self.problem.num_fixed_effects]
        gamma = x[self.problem.num_fixed_effects:self.problem.num_fixed_effects + self.problem.num_random_effects]
        return beta, gamma

    def joint_loss(self, x):
        beta, gamma = self.x_to_beta_gamma(x)
        return self.loss(beta, gamma)

    def joint_gradient(self, x):
        beta, gamma = self.x_to_beta_gamma(x)
        gradient = np.zeros(len(x))
        gradient[:self.problem.num_fixed_effects] = self.gradient_beta(beta, gamma)
        gradient[self.problem.num_fixed_effects:self.problem.num_fixed_effects + self.problem.num_random_effects] = \
            self.gradient_gamma(beta, gamma)
        return gradient

    def optimal_beta(self, gamma: np.ndarray, _dont_solve_wrt_beta=False, **kwargs):
        """
        Returns beta (vector of estimations of fixed effects) which minimizes loss function for a fixed gamma.

        The algorithm for computing optimal beta is::

            kernel = ∑X_i^TΩ_iX_i

            tail = ∑X_i^TΩ_iY_i

            β = (kernel)^{-1}*tail

        It's available almost exclusively in linear models. In general one should use gradient_beta and do iterative
        minimization instead.

        Parameters
        ----------
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        _dont_solve_wrt_beta : bool, Optional
            If true, then it does not perform the outer matrix inversion and returns the (kernel, tail) instead.
            It's left here for the purposes of use in child classes where both the kernel and the tail should be
            adjusted to account for regularization.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
        beta: np.ndarray, shape = [n]
            Vector of optimal estimates of the fixed effects for given gamma.
        """
        self._recalculate_cholesky(gamma)
        kernel = 0
        tail = 0
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            Lx = L_inv.dot(x)
            kernel += Lx.T.dot(Lx)
            tail += Lx.T.dot(L_inv.dot(y))
        if _dont_solve_wrt_beta:
            return kernel, tail
        else:
            return np.linalg.solve(kernel, tail)

    def optimal_random_effects(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> np.ndarray:
        """
        Returns set of optimal random effects estimations for given beta and gamma.

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
        u : np.ndarray, shape = [m, k]
            Set of optimal random effects estimations for given beta and gamma

        """

        random_effects = []
        self._recalculate_cholesky(gamma)
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            xi = y - x.dot(beta)
            stds_inv_mat = np.diag(1 / stds)
            # If the variance of R.E. is 0 then the R.E. is 0, so we take it into account separately
            # to keep matrices invertible.
            mask = np.abs(gamma) > 1e-10
            z_masked = z[:, mask]
            gamma_masked = gamma[mask]
            u_nonzero = np.linalg.solve(np.diag(1 / gamma_masked) + z_masked.T.dot(stds_inv_mat).dot(z_masked),
                                        z_masked.T.dot(stds_inv_mat).dot(xi)
                                        )
            u = np.zeros(len(gamma))
            u[mask] = u_nonzero
            random_effects.append(u)
        return np.array(random_effects)

    def optimal_obs_std(self, beta, gamma, **kwargs):
        self._recalculate_cholesky(gamma)
        result = 0
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            r = y - x.dot(beta)
            result += sum(L_inv.dot(r) ** 2)
        return result / self.problem.num_obs

    def _jones2010n_eff(self):
        n_eff = 0
        for (x, y, z, stds), L in zip(self.problem, self.omega_cholesky):
            omega = L.dot(L.T)
            sigma = np.sqrt(np.diag(omega)).reshape((-1, 1))
            normalization = sigma.dot(sigma.T)
            c = omega / normalization
            n_eff += np.linalg.inv(c).sum()
        return n_eff

    def jones2010bic(self, beta, gamma, **kwargs):
        # From here
        # https://www.researchgate.net/publication/51536734_Bayesian_information_criterion_for_longitudinal_and_clustered_data
        self._recalculate_cholesky(gamma)
        return self.loss(beta, gamma, **kwargs) + (len(beta) + len(gamma)) * np.log(self._jones2010n_eff())

    def _hat_matrix(self, gamma):
        self._recalculate_cholesky(gamma)

        h_beta_kernel = 0
        h_beta_tail = []
        xs = []
        # Form the beta kernel and tail
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            Lx = L_inv.dot(x)
            h_beta_kernel += Lx.T.dot(Lx)
            h_beta_tail.append(Lx.T.dot(L_inv))
            xs.append(x)

        h_beta = np.concatenate([np.linalg.inv(h_beta_kernel).dot(tail) for tail in h_beta_tail], axis=1)
        xs = np.concatenate(xs, axis=0)
        h_beta = xs.dot(h_beta)

        # we treat R.E. with very small variance
        # as effectively no R.E. to improve the stability of matrix inversions
        mask = np.abs(gamma) > 1e-10

        random_effects_parts = []
        for (x, y, z, stds), L_inv in zip(self.problem, self.omega_cholesky_inv):
            # Form the h_gamma
            stds_inv_mat = np.diag(1 / stds)
            # If the variance of R.E. is 0 then the R.E. is 0, so we take it into account separately
            # to keep matrices invertible.
            z_masked = z[:, mask]
            gamma_masked = gamma[mask]
            h_gamma_kernel = np.diag(1 / gamma_masked) + z_masked.T.dot(stds_inv_mat).dot(z_masked)
            h_gamma_tail = z_masked.T.dot(stds_inv_mat)
            h_gamma_i = np.zeros((self.problem.num_random_effects, z.shape[0]))
            h_gamma_i[mask, :] = np.linalg.inv(h_gamma_kernel).dot(h_gamma_tail)
            random_effects_parts.append(z.dot(h_gamma_i))

        h_gamma = sp.linalg.block_diag(*random_effects_parts)
        h = h_beta + h_gamma - h_gamma.dot(h_beta)
        return h

    def _hodges2001ddf(self, gamma, **kwargs):
        # From here:
        # https://www.jstor.org/stable/2673485?seq=1

        h_matrix = self._hat_matrix(gamma)
        return np.trace(h_matrix)

    def vaida2005aic(self, beta, gamma, **kwargs):
        # From here
        # https://www.jstor.org/stable/2673485?seq=1
        rho = self._hodges2001ddf(gamma)
        n = self.problem.num_obs
        p = sum(beta != 0)
        q = sum(gamma != 0)
        alpha = 2 * n / (n - p - 2) * (rho - (rho - p) / (n - p))
        # The likelihood here is conditional in the original paper
        # i.e. L(beta, gamma, us), but I put marginalized likelihood instead.
        return 2 * self.loss(beta, gamma, **kwargs) + alpha * (p + q)

    def get_ic(self, ic, beta, gamma, **kwargs):
        if ic == "IC_vaida2005aic":
            return self.vaida2005aic(beta, gamma, **kwargs)
        elif ic == "IC_jones2010bic":
            return self.jones2010bic(beta, gamma, **kwargs)
        else:
            raise ValueError(f"Unknown information criterion: {ic}")


class LinearLMEOracleRegularized(LinearLMEOracle):
    """
    Implements Regularized Linear Mixed-Effects Model functional for given problem::

        Y_i = X_i*β + Z_i*u_i + 𝜺_i,

        where

        β ~ 𝒩(tb, 1/lb),

        ||tβ||_0 = nnz(β) <= nnz_tbeta,

        u_i ~ 𝒩(0, diag(𝛄)),

        𝛄 ~ 𝒩(t𝛄, 1/lg),

        ||t𝛄||_0 = nnz(t𝛄) <= nnz_tgamma,

        𝜺_i ~ 𝒩(0, Λ)

    Here tβ and t𝛄 are single variables, not multiplications (e.g. not t*β). This oracle is designed for
    a solver (LinearLMESparseModel) which searches for a sparse solution (tβ, t𝛄) with at most k and j <= k non-zero
    elements respectively. For more details, see the documentation for LinearLMESparseModel.

    The problem should be provided as LinearLMEProblem.

    """

    def __init__(self, problem: LinearLMEProblem, lb=0.1, lg=0.1, nnz_tbeta=3, nnz_tgamma=3,
                 participation_in_selection=None):
        """
        Creates an oracle on top of the given problem. The problem should be in the form of LinearLMEProblem.

        Parameters
        ----------
        problem: LinearLMEProblem
            The set of data and answers. See the docs for LinearLMEProblem for more details.
        lb : float
            Regularization coefficient (inverse std) for ||β - tβ||^2
        lg : float
            Regularization coefficient (inverse std) for ||𝛄 - t𝛄||^2
        nnz_tbeta : int
            Number of non-zero elements allowed in tβ
        nnz_tgamma : int
            Number of non-zero elements allowed in t𝛄
        participation_in_selection : Tuple of Int, Optional, default = None
            Which features participate in selection. Defaults to None, which means all features participate in
            selection process
        """

        super().__init__(problem)
        self.lb = lb
        self.lg = lg
        self.k = nnz_tbeta
        self.j = nnz_tgamma
        self.participation_in_selection = participation_in_selection

    def optimal_beta(self, gamma: np.ndarray, tbeta: np.ndarray = None, _dont_solve_wrt_beta=False, **kwargs):
        """
        Returns beta (vector of estimations of fixed effects) which minimizes loss function for a fixed gamma.

        The algorithm for computing optimal beta is::

            kernel = ∑X_i^TΩ_iX_i

            tail = ∑X_i^TΩ_iY_i

            β = (kernel + I*lb)^{-1}*(tail + lb*tbeta)

        It's available almost exclusively in linear models. In general one should use gradient_beta and do iterative
        minimization instead.

        Parameters
        ----------
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        tbeta : np.ndarray, shape = [n]
            Vector of (nnz_tbeta)-sparse estimates for fixed effects.
        _dont_solve_wrt_beta : bool, Optional
            If true, then it does not perform the outer matrix inversion and returns the (kernel, tail) instead.
            It's left here for the purposes of use in child classes where both the kernel and the tail should be
            adjusted to account for regularization.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
        beta: np.ndarray, shape = [n]
            Vector of optimal estimates of the fixed effects for given gamma.
        """
        kernel, tail = super().optimal_beta(gamma, _dont_solve_wrt_beta=True, **kwargs)
        if _dont_solve_wrt_beta:
            return kernel, tail
        return np.linalg.solve(self.lb * np.eye(self.problem.num_fixed_effects) + kernel, self.lb * tbeta + tail)

    @staticmethod
    def _take_only_k_max(x: np.ndarray, k: int, **kwargs):
        """
        Returns a vector b which consists of k largest elements of x (at the same places) and zeros everywhere else.

        Parameters
        ----------
        x : np.ndarray, shape = [n]
            Vector which we take largest elements from.
        k : int
            How many elements we take from x
        kwargs :
            Not used, left for future and for passing debug/experimental parameters.

        Returns
        -------
        b : np.ndarray, shape = [n]
            A vector which consists of k largest elements of x (at the same places) and zeros everywhere else.
        """

        b = np.zeros(len(x))
        if k > 0:
            idx_k_max = np.abs(x).argsort()[-k:]
            b[idx_k_max] = x[idx_k_max]
        return b

    def optimal_tbeta(self, beta: np.ndarray, **kwargs):
        """
        Returns tbeta which minimizes the loss function with all other variables fixed.

        It is a projection of beta on the sparse subspace with no more than k elements, which can be constructed by
        taking k largest elements from beta and setting the rest to be 0.

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters.

        Returns
        -------
        tbeta : np.ndarray, shape = [n]
            Minimizer of the loss function w.r.t tbeta with other arguments fixed.
        """
        if self.participation_in_selection is not None:
            result = np.copy(beta)
            result[self.participation_in_selection] = self._take_only_k_max(beta[self.participation_in_selection],
                                                                            self.k - sum(
                                                                                ~self.participation_in_selection))
            return result
        else:
            return self._take_only_k_max(beta, self.k, **kwargs)

    def loss(self, beta: np.ndarray, gamma: np.ndarray, tbeta: np.ndarray = None, tgamma: np.ndarray = None, **kwargs):
        """
        Returns the loss function value ℒ(β, 𝛄) + lb/2*||β - tβ||^2 + lg/2*||𝛄 - t𝛄||^2

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        tbeta : np.ndarray, shape = [n]
            Vector of (nnz_tbeta)-sparse estimates of fixed effects.
        tgamma : np.ndarray, shape = [k]
            Vector of (nnz_tgamma)-sparse estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters.

        Returns
        -------
            result : float
                The value of the loss function: ℒ(β, 𝛄) + lb/2*||β - tβ||^2 + lg/2*||𝛄 - t𝛄||^2
        """

        return (super().loss(beta, gamma, **kwargs)
                + self.lb / 2 * sum((beta - tbeta) ** 2)
                + self.lg / 2 * sum((gamma - tgamma) ** 2))

    def gradient_gamma(self, beta: np.ndarray, gamma: np.ndarray, tgamma: np.ndarray = None, **kwargs) -> np.ndarray:
        """
        Returns the gradient of the loss function with respect to gamma: grad_gamma =  ∇_𝛄[ℒ](β, 𝛄) + lg*(𝛄 - t𝛄)

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        tgamma : np.ndarray, shape = [k]
            Vector of (nnz_tgamma)-sparse covariance estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
            grad_gamma: np.ndarray, shape = [k]
                The gradient of the loss function with respect to gamma: grad_gamma = ∇_𝛄[ℒ](β, 𝛄) + lg*(𝛄 - t𝛄)
        """

        return super().gradient_gamma(beta, gamma, **kwargs) + self.lg * (gamma - tgamma)

    def hessian_gamma(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> np.ndarray:
        """
        Returns the Hessian of the loss function with respect to gamma: ∇²_𝛄[ℒ](β, 𝛄) + lg*I.

        Parameters
        ----------
        beta : np.ndarray, shape = [n]
            Vector of estimates of fixed effects.
        gamma : np.ndarray, shape = [k]
            Vector of estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters

        Returns
        -------
            hessian: np.ndarray, shape = [k, k]
                Hessian of the loss function with respect to gamma ∇²_𝛄[ℒ](β, 𝛄) + lg*I.
        """

        return super().hessian_gamma(beta, gamma, **kwargs) + self.lg * np.eye(self.problem.num_random_effects)

    def optimal_tgamma(self, tbeta, gamma, **kwargs):
        """
        Returns tgamma which minimizes the loss function with all other variables fixed.

        It is a projection of gamma on the sparse subspace with no more than nnz_gamma elements,
        which can be constructed by taking nnz_gamma largest elements from gamma and setting the rest to be 0.
        In addition, it preserves that for all the elements where tbeta = 0 it implies that tgamma = 0 as well.

        Parameters
        ----------
        tbeta : np.ndarray, shape = [n]
            Vector of (nnz_beta)-sparse estimates for fixed parameters.
        gamma : np.ndarray, shape = [k]
            Vector of covariance estimates of random effects.
        kwargs :
            Not used, left for future and for passing debug/experimental parameters.

        Returns
        -------
        tgamma : np.ndarray, shape = [k]
            Minimizer of the loss function w.r.t tgamma with other arguments fixed.
        """
        tgamma = np.zeros(len(gamma))
        idx = tbeta != 0
        idx_gamma = self.beta_to_gamma_map[idx]
        idx_gamma = (idx_gamma[idx_gamma >= 0]).astype(int)
        tgamma[idx_gamma] = gamma[idx_gamma]
        if self.participation_in_selection is not None:
            participation_idx = self.beta_to_gamma_map[self.participation_in_selection]
            participation_idx = (participation_idx[participation_idx >= 0]).astype(int)
            # if tbeta = 0 then tgamma = 0 even if this coordinate does not participate in feature selection
            not_participation_idx = self.beta_to_gamma_map[~self.participation_in_selection & (tbeta != 0)]
            not_participation_idx = (not_participation_idx[not_participation_idx >= 0]).astype(int)
            tgamma[not_participation_idx] = gamma[not_participation_idx]
            tgamma[participation_idx] = self._take_only_k_max(tgamma[participation_idx],
                                                              self.j - sum(~self.participation_in_selection))
            return tgamma
        else:
            return self._take_only_k_max(tgamma, self.j)


class LinearLMEOracleW(LinearLMEOracleRegularized):

    def __init__(self, problem: LinearLMEProblem, lb=0.1, lg=0.1, nnz_tbeta=3, nnz_tgamma=3):
        super().__init__(problem, lb, lg, nnz_tbeta, nnz_tgamma)
        self.beta = None
        self.drop_penalties_beta = None
        self.drop_penalties_gamma = None

    def _recalculate_drop_matrices(self, beta, gamma):
        if np.all(self.gamma == gamma) and np.all(self.beta == beta):
            return None
        self._recalculate_cholesky(gamma)

        self.drop_penalties_beta = np.zeros(self.problem.num_fixed_effects)
        self.drop_penalties_gamma = np.zeros(self.problem.num_random_effects)
        for j, ((x, y, z, l), L_inv) in enumerate(zip(self.problem, self.omega_cholesky_inv)):
            # Calculate drop price for gammas individually
            xi = y - x.dot(beta)
            Lxi = L_inv.dot(xi)
            Lx = L_inv.dot(x)
            Lz = L_inv.dot(z)
            h1 = np.sum(Lz ** 2, axis=0)
            g1 = Lz.T.dot(Lxi) ** 2
            self.drop_penalties_gamma += -gamma * g1 / (1 - gamma * h1) + np.log(1 + gamma * h1 / (1 - gamma * h1))
            # Calculate drop price for betas only
            self.drop_penalties_beta += -2 * beta * Lx.T.dot(Lxi)
            self.drop_penalties_beta += -beta ** 2 * np.sum(Lx ** 2, axis=0)
            # Calculate drop price for gammas given dropped betas
            idx_beta = []
            idx_gamma = []
            for i, k in enumerate(self.beta_to_gamma_map):
                if k >= 0:
                    idx_beta.append(i)
                    idx_gamma.append(k)
            idx_beta = np.array(idx_beta)
            idx_gamma = np.array(idx_gamma).astype(int)
            x_s = x[:, idx_beta]
            beta_s = beta[idx_beta]
            gamma_s = gamma[idx_gamma]
            z_s = z[:, idx_gamma]
            Lz_s = L_inv.dot(z_s)
            h1_s = np.sum(Lz_s ** 2, axis=0)
            g2_s = np.sum((Lxi.reshape((len(Lxi), 1)) + L_inv.dot(x_s * beta_s)) * Lz_s, axis=0)
            self.drop_penalties_beta[idx_beta] += (-gamma_s * (g2_s ** 2) / (1 - gamma_s * h1_s)
                                                   + np.log(1 + gamma_s * h1_s / (1 - gamma_s * h1_s)))

        # we invert the sign and take into account the 1/2 multiplier for the loss function
        self.drop_penalties_beta /= -2
        self.drop_penalties_gamma /= -2
        self.beta = beta
        return None

    def loss(self, beta: np.ndarray, gamma: np.ndarray, tbeta: np.ndarray = None, tgamma: np.ndarray = None, **kwargs):
        if self.drop_penalties_beta is None or self.drop_penalties_gamma is None:
            self._recalculate_drop_matrices(beta, gamma)
        return (super(LinearLMEOracleRegularized, self).loss(beta, gamma, **kwargs)
                + self.lb / 2 * sum(self.drop_penalties_beta * (beta - tbeta) ** 2)
                + self.lg / 2 * sum(self.drop_penalties_gamma * (gamma - tgamma) ** 2))

    def optimal_beta(self, gamma: np.ndarray, tbeta: np.ndarray = None, beta: np.ndarray = None, **kwargs):
        if self.drop_penalties_beta is None:
            if beta is not None:
                self._recalculate_drop_matrices(beta, gamma)
            else:
                raise ValueError("Drop penalties for beta are not initialized")

        kernel, tail = super(LinearLMEOracleRegularized, self).optimal_beta(gamma, _dont_solve_wrt_beta=True, **kwargs)
        return np.linalg.solve(self.lb * np.diag(self.drop_penalties_beta) + kernel,
                               self.lb * self.drop_penalties_beta * tbeta + tail)

    def gradient_gamma(self, beta: np.ndarray, gamma: np.ndarray, tgamma: np.ndarray = None, **kwargs) -> np.ndarray:
        if self.drop_penalties_gamma is None:
            self._recalculate_drop_matrices(beta, gamma)
        return (super(LinearLMEOracleRegularized, self).gradient_gamma(beta, gamma, tgamma=tgamma, **kwargs)
                + self.lg * self.drop_penalties_gamma * (gamma - tgamma))

    def hessian_gamma(self, beta: np.ndarray, gamma: np.ndarray, **kwargs) -> np.ndarray:
        if self.drop_penalties_gamma is None:
            self._recalculate_drop_matrices(beta, gamma)
        return super(LinearLMEOracleRegularized, self).hessian_gamma(beta, gamma, **kwargs) + self.lg * np.diag(
            self.drop_penalties_gamma)

    def optimal_tbeta(self, beta: np.ndarray, gamma: np.ndarray = None, **kwargs):
        self._recalculate_drop_matrices(beta, gamma)
        tbeta = np.zeros(len(beta))
        idx_k_max = np.abs(self.drop_penalties_beta * beta).argsort()[-self.k:]
        tbeta[idx_k_max] = beta[idx_k_max]
        return tbeta

    def optimal_tgamma(self, tbeta, gamma, beta: np.ndarray = None, **kwargs):
        self._recalculate_drop_matrices(beta, gamma)
        tgamma = np.zeros(len(gamma))
        idx = tbeta != 0
        idx_gamma = self.beta_to_gamma_map[idx]
        idx_gamma = (idx_gamma[idx_gamma >= 0]).astype(int)
        tgamma[idx_gamma] = gamma[idx_gamma]
        idx_k_max = np.abs(self.drop_penalties_gamma * tgamma).argsort()[-self.j:]
        tgamma2 = np.zeros(len(gamma))
        tgamma2[idx_k_max] = tgamma[idx_k_max]
        return tgamma2

    def jones2010bic(self, beta, gamma, tbeta=None, tgamma=None, **kwargs):
        self._recalculate_cholesky(gamma)
        return self.loss(beta, gamma, tbeta, tgamma) + (2 * len(beta) + 2 * len(gamma)) * np.log(self._jones2010n_eff())


class LinearLMELassoOracle(LinearLMEOracle):
    def __init__(self, problem: LinearLMEProblem, lb=0.1, lg=0.1):
        """
        Creates an oracle on top of the given problem. The problem should be in the form of LinearLMEProblem.

        Parameters
        ----------
        problem: LinearLMEProblem
            The set of data and answers. See the docs for LinearLMEProblem for more details.
        lb : float
            Regularization coefficient (inverse std) for ||β||_1
        lg : float
            Regularization coefficient (inverse std) for ||𝛄||_1
        """

        super().__init__(problem)
        self.lb = lb
        self.lg = lg
        lambdas = np.zeros(problem.num_fixed_effects+problem.num_random_effects)
        lambdas[:problem.num_fixed_effects] = lb
        lambdas[problem.num_fixed_effects:problem.num_fixed_effects+problem.num_random_effects] = lg
        self.lambdas = lambdas

    def full_loss(self, x):
        return self.joint_loss(x) + self.lambdas.dot(abs(x))

    def prox_l1(self, x, step_len):
        """
        Applies proximal operator of l1 norm to x and returns result

        Parameters
        ----------
        x : np.ndarray
            vector to apply the proximal operator on
        """
        return np.sign(x)*np.maximum(0.0, np.abs(x) - self.lambdas*step_len)



# This code implements different skmixed's subroutines.
# Copyright (C) 2020 Aleksei Sholokhov, aksh@uw.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np


def get_per_group_coefficients(beta, random_effects, labels):
    """
    Derives per group coefficients from the vectors of fixed and per-cluster random effects.

    Parameters
    ----------
    beta: np.ndarray, shape=(n,), n is the number of fixed effects.
        Vector of fixed effects.
    random_effects: np.ndarray or list, shape=(m, k), m groups, k random effects.
        Array of random effects.
    labels: np.ndarray[int], shape=(t,), t -- number of columns in the dataset INCLUDING INTERCEPT.
        Vector of labels of the column's dataset, including intercept. Labels can be the following integers:
            0 : Groups labels (ignored).
            1 : Fixed effect.
            2 : Random effect.
            3 : Both fixed and random effect.
            4 : Standard deviations for measurement errors for answers (ignored).

    Returns
    -------
    per_group_coefficients: np.ndarray, shape=(m, t)
        Array of cluster coefficients: m clusters times t coefficients.
    """
    is_arrays = False
    if all([type(s) == np.ndarray for s in random_effects]):
        random_effects = [(None, u) for u in random_effects]
        is_arrays = True

    per_group_coefficients_list = []

    for i, (label, u) in enumerate(random_effects):
        per_group_coefficients = np.zeros(len(labels))
        fixed_effects_counter = 0
        random_effects_counter = 0

        for j, feature_label in enumerate(labels):
            if feature_label == 1:
                per_group_coefficients[j] = beta[fixed_effects_counter]
                fixed_effects_counter += 1
            elif feature_label == 2:
                per_group_coefficients[j] = u[random_effects_counter]
                random_effects_counter += 1
            elif feature_label == 3:
                per_group_coefficients[j] = beta[fixed_effects_counter] + u[random_effects_counter]
                fixed_effects_counter += 1
                random_effects_counter += 1
            else:
                continue

        per_group_coefficients_list.append((label, per_group_coefficients))
    if is_arrays:
        return random_effects_to_matrix(per_group_coefficients_list)
    else:
        return per_group_coefficients_list


def random_effects_to_matrix(random_effects):
    return np.array([u for k, u in random_effects])



import unittest

import numpy as np
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LinearLMESparseModel
from skmixed.lme.problems import LinearLMEProblem

from skmixed.helpers import random_effects_to_matrix


class TestLinearLMESparseModel(unittest.TestCase):

    def test_solving_dense_problem(self):
        trials = 20
        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [3, 3, 3],
            "random_intercept": True,
            "features_covariance_matrix": np.array([
                [1, 0, 0],
                [0, 1, 0.7],
                [0, 0.7, 1]
            ]),
            "obs_std": 0.1,
        }
        model_parameters = {
            "nnz_tbeta": 2,
            "nnz_tgamma": 2,
            "lb": 0,        # We expect the coefficient vectors to be dense so we turn regularization off.
            "lg": 0,        # Same.
            "initializer": 'EM',
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
            "n_iter_outer": 1  # we don't care about tbeta and tgamma, so we don't increase regularization iteratively
        }

        max_mse = 0.05
        min_explained_variance = 0.9

        for i in range(trials):
            problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters, seed=i)
            model = LinearLMESparseModel(**model_parameters)

            x, y = problem.to_x_y()
            # model.fit(x, y)
            model.fit_problem(problem)

            logger = model.logger_
            loss = np.array(logger.get("loss"))
            self.assertTrue(np.all(loss[1:] - loss[:-1] <= 0),
                            msg="%d) Loss does not decrease monotonically with iterations. (seed=%d)" % (i, i))

            y_pred = model.predict_problem(problem)
            explained_variance = explained_variance_score(y, y_pred)
            mse = mean_squared_error(y, y_pred)

            # coefficients = model.coef_
            # maybe_per_group_coefficients = coefficients["per_group_coefficients"]

            self.assertGreater(explained_variance, min_explained_variance,
                               msg="%d) Explained variance is too small: %.3f < %.3f. (seed=%d)"
                                   % (i,
                                      explained_variance,
                                      min_explained_variance,
                                      i))
            self.assertGreater(max_mse, mse,
                               msg="%d) MSE is too big: %.3f > %.2f  (seed=%d)"
                                   % (i,
                                      mse,
                                      max_mse,
                                      i))

            # coefficients = model.coef_
            # maybe_per_group_coefficients = coefficients["per_group_coefficients"]
            # maybe_beta = coefficients["beta"]
            # maybe_us = coefficients["random_effects"]
            # maybe_gamma = coefficients["gamma"]
            # maybe_tbeta = coefficients["tbeta"]
            # maybe_tgamma = coefficients["tgamma"]
            # maybe_cluster_coefficients = coefficients["per_cluster_coefficients"]
            # maybe_sparse_cluster_coefficients = coefficients["sparse_per_cluster_coefficients"]
        # cluster_coefficients = beta + us
        # maybe_cluster_coefficients = maybe_beta + maybe_us
        return None

    def test_solving_sparse_problem(self):
        trials = 10
        problem_parameters = {
            "groups_sizes": [20, 12, 14, 50, 11],
            "features_labels": [3, 3, 3, 3, 3, 3, 3, 3, 3, 3],
            "random_intercept": True,
            "obs_std": 0.1,
        }

        model_parameters = {
            # "nnz_tbeta": 4,  # we define them later in trial iterations
            # "nnz_tgamma": 3,
            "lb": 20,
            "lg": 2,
            "initializer": None,
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
            "n_iter_outer": 20
        }

        max_mse = 0.1
        min_explained_variance = 0.9
        fixed_effects_min_accuracy = 0.7
        random_effects_min_accuracy = 0.7

        for i in range(trials):
            np.random.seed(i)
            true_beta = np.random.choice(2, size=11, p=np.array([0.5, 0.5]))
            if sum(true_beta) == 0:
                true_beta[0] = 1
            true_gamma = np.random.choice(2, size=11, p=np.array([0.3, 0.7]))*true_beta

            problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters,
                                                                       beta=true_beta,
                                                                       gamma=true_gamma,
                                                                       seed=i)
            model = LinearLMESparseModel(**model_parameters,
                                         nnz_tbeta=sum(true_beta),
                                         nnz_tgamma=sum(true_gamma))

            x, y = problem.to_x_y()
            # model.fit(x, y)
            model.fit_problem(problem)

            logger = model.logger_
            loss = np.array(logger.get("loss"))
            # TODO: It won't decrease monotonically because it may jump when we increase regularization.
            # self.assertTrue(np.all(loss[1:] - loss[:-1] <= 0),
            #                 msg="%d) Loss does not decrease monotonically with iterations. (seed=%d)" % (i, i))

            y_pred = model.predict_problem(problem)
            explained_variance = explained_variance_score(y, y_pred)
            mse = mean_squared_error(y, y_pred)

            coefficients = model.coef_
            maybe_tbeta = coefficients["tbeta"]
            maybe_tgamma = coefficients["tgamma"]
            fixed_effects_accuracy = accuracy_score(true_beta, maybe_tbeta != 0)
            random_effects_accuracy = accuracy_score(true_gamma, maybe_tgamma != 0)

            # maybe_per_group_coefficients = coefficients["per_group_coefficients"]

            self.assertGreater(explained_variance, min_explained_variance,
                               msg="%d) Explained variance is too small: %.3f < %.3f. (seed=%d)"
                                   % (i,
                                      explained_variance,
                                      min_explained_variance,
                                      i))
            self.assertGreater(max_mse, mse,
                               msg="%d) MSE is too big: %.3f > %.2f  (seed=%d)"
                                   % (i,
                                      mse,
                                      max_mse,
                                      i))
            self.assertGreater(fixed_effects_accuracy, fixed_effects_min_accuracy,
                               msg="%d) Fixed Effects Selection Accuracy is too small: %.3f < %.2f  (seed=%d)"
                                   % (i,
                                      fixed_effects_accuracy,
                                      fixed_effects_min_accuracy,
                                      i)
                               )
            self.assertGreater(random_effects_accuracy, random_effects_min_accuracy,
                               msg="%d) Random Effects Selection Accuracy is too small: %.3f < %.2f  (seed=%d)"
                                   % (i,
                                      random_effects_accuracy,
                                      random_effects_min_accuracy,
                                      i)
                               )
        return None

    def test_get_set_params(self):
        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [3, 3, 3],
            "random_intercept": True,
            "features_covariance_matrix": np.array([
                [1, 0, 0],
                [0, 1, 0.7],
                [0, 0.7, 1]
            ]),
            "obs_std": 0.1,
        }
        model_parameters = {
            "nnz_tbeta": 4,
            "nnz_tgamma": 4,
            "lb": 0,  # We expect the coefficient vectors to be dense so we turn regularization off.
            "lg": 0,  # Same.
            "initializer": 'EM',
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
        }
        # Now we want to solve a regularized problem to get two different models
        model2_parameters = {
            "nnz_tbeta": 3,
            "nnz_tgamma": 2,
            "lb": 20,
            "lg": 20,
            "initializer": None,
            "logger_keys": ('converged',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
        }
        problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters, seed=42)
        x, y = problem.to_x_y()

        model = LinearLMESparseModel(**model_parameters)
        # model.fit(x, y)
        model.fit_problem(problem)
        params = model.get_params()
        y_pred = model.predict_problem(problem)

        model2 = LinearLMESparseModel(**model2_parameters)
        #model2.fit(x, y)
        model2.fit_problem(problem)
        params2 = model2.get_params()
        y_pred2 = model2.predict_problem(problem)

        model.set_params(**params2)
        model.fit_problem(problem)
        y_pred_with_other_params = model.predict_problem(problem)
        assert np.equal(y_pred_with_other_params, y_pred2).all(),\
            "set_params or get_params is not working properly"
        model2.set_params(**params)
        model2.fit_problem(problem)
        y_pred2_with_other_params = model2.predict_problem(problem)
        assert np.equal(y_pred2_with_other_params, y_pred).all(), \
            "set_params or get_params is not working properly"

    def test_score_function(self):
        # this is only a basic test which checks R^2 in two points: nearly perfect prediction and constant prediction.

        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [3, 3, 3],
            "random_intercept": True,
            "features_covariance_matrix": np.array([
                [1, 0, 0],
                [0, 1, 0.7],
                [0, 0.7, 1]
            ]),
            "obs_std": 0.1,
        }

        model_parameters = {
            "nnz_tbeta": 4,
            "nnz_tgamma": 4,
            "lb": 0,  # We expect the coefficient vectors to be dense so we turn regularization off.
            "lg": 0,  # Same.
            "initializer": 'EM',
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-6,
            "n_iter": 1000,
            "tol_inner": 1e-4,
            "n_iter_inner": 1000,
        }

        problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters, seed=42)
        x, y = problem.to_x_y()
        model = LinearLMESparseModel(**model_parameters)
        model.fit_problem(problem)
        model.coef_["beta"] = true_model_parameters["beta"]
        model.coef_["random_effects"] = random_effects_to_matrix(true_model_parameters["random_effects"])
        good_score = model.score(x, y)
        assert good_score > 0.99
        model.coef_["beta"] = np.zeros(4)
        model.coef_["random_effects"] = np.zeros((4, 4))
        bad_score = model.score(x, y)
        assert abs(bad_score) < 0.1


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from sklearn.metrics import mean_squared_error, explained_variance_score, accuracy_score

from skmixed.lme.models import LassoLMEModel
from skmixed.lme.problems import LinearLMEProblem


class TestLassoLMEModel(unittest.TestCase):

    def test_solving_without_regularization(self):
        trials = 20
        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [3, 3, 3],
            "random_intercept": True,
            "features_covariance_matrix": np.array([
                [1, 0, 0],
                [0, 1, 0.7],
                [0, 0.7, 1]
            ]),
            "obs_std": 0.1,
        }
        model_parameters = {
            "lb": 0,        # We expect the coefficient vectors to be dense so we turn regularization off.
            "lg": 0,        # Same.
            "logger_keys": ('converged', 'loss',),
            "tol": 1e-3,
            "n_iter": 1000,
        }

        max_mse = 0.05
        min_explained_variance = 0.9

        for i in range(trials):
            problem, true_model_parameters = LinearLMEProblem.generate(**problem_parameters, seed=i)
            x, y = problem.to_x_y()

            model = LassoLMEModel(**model_parameters)
            model.fit_problem(problem)

            logger = model.logger_
            loss = np.array(logger.get("loss"))
            self.assertTrue(np.all(loss[1:] - loss[:-1] <= 0),
                            msg="%d) Loss does not decrease monotonically with iterations. (seed=%d)" % (i, i))

            y_pred = model.predict_problem(problem)
            explained_variance = explained_variance_score(y, y_pred)
            mse = mean_squared_error(y, y_pred)

            self.assertGreater(explained_variance, min_explained_variance,
                               msg="%d) Explained variance is too small: %.3f < %.3f. (seed=%d)"
                                   % (i,
                                      explained_variance,
                                      min_explained_variance,
                                      i))
            self.assertGreater(max_mse, mse,
                               msg="%d) MSE is too big: %.3f > %.2f  (seed=%d)"
                                   % (i,
                                      mse,
                                      max_mse,
                                      i))

        return None


import numpy as np
from scipy.optimize import minimize
from scipy.linalg.lapack import get_lapack_funcs

from skmixed.lme.problems import LinearLMEProblem


class LinearLMEOracle:
    def __init__(self, problem: LinearLMEProblem, mode='fast'):
        self.problem = problem
        assert mode == 'naive' or mode == 'fast', "Unknown mode: %s" % mode
        self.mode = mode
        if self.mode == 'fast':
            self.omegas_inv = []
            self.zTomegas_inv = []
            self.zTomegas_invZ = []
            self.xTomegas_invY = []
            self.xTomegas_invX = []
            self.old_gamma = None

    def recalculate_inverse_matrices(self, gamma: np.ndarray) -> None:
        if (self.old_gamma == gamma).all():
            return None
        if self.old_gamma is None:
            self.omegas_inv = []
            self.zTomegas_inv = []
            self.zTomegas_invZ = []
            self.xTomegas_invY = []
            self.xTomegas_invX = []
            for x, y, z, l in self.problem:
                l_mat = np.diag(l)
                omega_inv = np.linalg.inv(z.dot(np.diag(gamma)).dot(z.T) + l_mat)
                zTomega = z.T.dot(omega_inv)
                zTomegaZ = zTomega.dot(z)
                self.omegas_inv.append(omega_inv)
                self.zTomegas_inv.append(zTomega)
                self.zTomegas_invZ.append(zTomegaZ)
                xTomega_inv = x.T.dot(omega_inv)
                self.xTomegas_invX.append(xTomega_inv.dot(x))
                self.xTomegas_invY.append(xTomega_inv.dot(y))
            self.old_gamma = gamma
        else:
            if (self.old_gamma - gamma == 0).any():
                self.old_gamma = None
                self.recalculate_inverse_matrices(gamma)
            else:
                dGamma_inv = np.diag(1 / (self.old_gamma - gamma))
                for i, (x, y, z, l) in enumerate(self.problem):
                    kernel_update = np.linalg.inv(dGamma_inv - self.zTomegas_invZ[i])
                    new_omega = self.omegas_inv[i] + self.omegas_inv[i].dot(z).dot(kernel_update).dot(
                        self.zTomegas_inv[i])
                    new_zTomega = z.T.dot(new_omega)
                    new_zTomegaZ = new_zTomega.dot(z)
                    self.omegas_inv[i] = new_omega
                    self.zTomegas_inv[i] = new_zTomega
                    self.zTomegas_invZ[i] = new_zTomegaZ
                    xTomega_inv = x.T.dot(new_omega)
                    self.xTomegas_invY[i] = xTomega_inv.dot(y)
                    self.xTomegas_invX[i] = xTomega_inv.dot(x)
                self.old_gamma = gamma

    def loss(self, beta: np.ndarray, gamma: np.ndarray) -> float:
        gamma_mat = np.diag(gamma)
        result = 0
        problem = self.problem
        if self.mode == 'naive':
            for x, y, z, l in problem:
                omega = z.dot(gamma_mat).dot(z.T) + np.diag(l)
                xi = y - x.dot(beta)
                sign, determinant = np.linalg.slogdet(omega)
                result += 1 / 2 * xi.T.dot(np.linalg.inv(omega)).dot(xi) + 1 / 2 * sign * determinant
        elif self.mode == 'fast':
            self.recalculate_inverse_matrices(gamma)
            for i, (x, y, z, l) in enumerate(problem):
                omega_inv = self.omegas_inv[i]
                xi = y - x.dot(beta)
                sign, determinant = np.linalg.slogdet(omega_inv)
                # Minus because we need the det of omega but use the det of its inverse
                result += 1 / 2 * xi.T.dot(omega_inv).dot(xi) - 1 / 2 * sign * determinant
        else:
            raise Exception("Unknown mode: %s" % self.mode)
        return result

    def gradient_gamma(self, beta: np.ndarray, gamma: np.ndarray) -> np.ndarray:
        if self.mode == 'naive':
            gamma_mat = np.diag(gamma)
            grad_gamma = np.zeros(len(gamma))
            for j in range(len(gamma)):
                result = 0
                for x, y, z, l in self.problem:
                    omega_inv = z.dot(gamma_mat).dot(z.T) + np.diag(l)
                    xi = y - x.dot(beta)
                    z_col = z[:, j]
                    data_part = z_col.T.dot(np.linalg.inv(omega_inv)).dot(xi)
                    data_part = -1 / 2 * data_part ** 2
                    det_part = 1 / 2 * z_col.T.dot(np.linalg.inv(omega_inv)).dot(z_col)
                    result += data_part + det_part
                grad_gamma[j] = result
            return grad_gamma
        elif self.mode == 'fast':
            if (self.old_gamma == gamma).all():
                result = np.zeros(self.problem.num_random_effects)
                for i, (x, y, z, l) in enumerate(self.problem):
                    xi = y - x.dot(beta)
                    result += 1 / 2 * (np.diag(self.zTomegas_invZ[i]) - self.zTomegas_inv[i].dot(xi) ** 2)
                self.old_gamma = gamma
                return result
            else:
                self.recalculate_inverse_matrices(gamma)
                result = np.zeros(self.problem.num_random_effects)
                for i, (x, y, z, l) in enumerate(self.problem):
                    new_zTomega = self.zTomegas_inv[i]
                    new_zTomegaZ = self.zTomegas_invZ[i]
                    xi = y - x.dot(beta)
                    result += 1 / 2 * (np.diag(new_zTomegaZ) - new_zTomega.dot(xi) ** 2)
                return result

    def hessian_gamma(self, beta: np.ndarray, gamma: np.ndarray) -> np.ndarray:
        if self.mode == 'naive':
            raise NotImplementedError(
                "Hessians are not implemented for the naive mode (it takes forever to compute them)")
        elif self.mode == 'fast':
            result = np.zeros((self.problem.num_random_effects, self.problem.num_random_effects))
            self.recalculate_inverse_matrices(gamma)
            for i, (x, y, z, l) in enumerate(self.problem):
                xi = y - x.dot(beta)
                eta = self.zTomegas_inv[i].dot(xi)
                eta = eta.reshape(len(eta), 1)
                result -= self.zTomegas_invZ[i] ** 2
                result += 2 * eta.dot(eta.T) * self.zTomegas_invZ[i]
            return 1 / 2 * result
        else:
            raise Exception("Unknown mode: %s" % self.mode)

    def optimal_beta(self, gamma: np.ndarray):  # , force_naive=False):
        omega = 0
        tail = 0

        if self.mode == 'naive':
            gamma_mat = np.diag(gamma)
            for x, y, z, l in self.problem:
                omega_i = z.dot(gamma_mat).dot(z.T) + np.diag(l)
                omega += x.T.dot(np.linalg.inv(omega_i)).dot(x)
                tail += x.T.dot(np.linalg.inv(omega_i)).dot(y)
        elif self.mode == 'fast':
            if not (self.old_gamma == gamma).all():
                self.recalculate_inverse_matrices(gamma)
            omega = np.sum(self.xTomegas_invX, axis=0)
            tail = np.sum(self.xTomegas_invY, axis=0)
        else:
            raise Exception("Unexpected mode: %s" % self.mode)
        return np.linalg.inv(omega).dot(tail)

    def optimal_random_effects(self, beta: np.ndarray, gamma: np.ndarray):
        random_effects = []
        for x, y, z, l in self.problem:
            l = np.diag(l)
            # This is an ad-hoc to make the matrix gamma invertible even when
            # some \gamma_i are zero. We fix it shortly after.
            # TODO: implement better account for zero gamma (need to del. resp. rows and columns from all the matrices)
            inv_g = np.diag(np.array([0 if g == 0 else 1 / g for g in gamma]))
            u = np.linalg.inv(inv_g + z.T.dot(np.linalg.inv(l)).dot(z)).dot(
                z.T.dot(np.linalg.inv(l)).dot(y - x.dot(beta)))
            # Here we put all the random effects for zero gammas to be zero
            for i, g in enumerate(gamma):
                if g == 0:
                    u[i] = 0
            random_effects.append(u)
        return np.array(random_effects)


class LinearLMEOracleRegularized(LinearLMEOracle):
    def __init__(self, problem: LinearLMEProblem, mode='fast', lb=0.1, lg=0.1, k=3, j=3):
        super().__init__(problem, mode)
        self.lb = lb
        self.lg = lg
        self.k = k
        self.j = j

    def optimal_beta_reg(self, gamma: np.ndarray, tbeta: np.ndarray):
        omega = 0
        tail = 0
        invert_upper_triangular = get_lapack_funcs("trtri")
        if self.mode == 'naive':
            gamma_mat = np.diag(gamma)
            for x, y, z, l in self.problem:
                omega_i = z.dot(gamma_mat).dot(z.T) + np.diag(l)
                L = np.linalg.cholesky(omega_i)
                #L_inv = invert_upper_triangular(L.T)[0].T
                #Lx = L_inv.dot(x)
                #omega += Lx.T.dot(Lx)
                omega += x.T.dot(np.linalg.inv(omega_i)).dot(x)
                #omega += np.diag(1/gamma)
                tail += x.T.dot(np.linalg.solve(omega_i, y))
        elif self.mode == 'fast':
            if not (self.old_gamma == gamma).all():
                self.recalculate_inverse_matrices(gamma)
            omega = np.sum(self.xTomegas_invX, axis=0)
            tail = np.sum(self.xTomegas_invY, axis=0)
        else:
            raise Exception("Unexpected mode: %s" % self.mode)
        return np.linalg.solve(self.lb * np.eye(self.problem.num_fixed_effects) + omega, self.lb * tbeta + tail)

    @staticmethod
    def take_only_k_max(a: np.ndarray, k: int):
        b = np.zeros(len(a))
        idx_k_max = a.argsort()[-k:]
        b[idx_k_max] = a[idx_k_max]
        return b

    def optimal_tbeta(self, beta: np.ndarray):
        return self.take_only_k_max(beta, self.k)

    def loss_reg(self, beta: np.ndarray, gamma: np.ndarray, tbeta: np.ndarray, tgamma: np.ndarray):
        return self.loss(beta, gamma) + self.lb / 2 * sum((beta - tbeta) ** 2) + \
               self.lg / 2 * sum((gamma - tgamma) ** 2)

    def gradient_gamma_reg(self, beta: np.ndarray, gamma: np.ndarray, tgamma: np.ndarray) -> np.ndarray:
        return self.gradient_gamma(beta, gamma) + self.lg * (gamma - tgamma)

    def hessian_gamma_reg(self, beta: np.ndarray, gamma: np.ndarray) -> np.ndarray:
        return self.hessian_gamma(beta, gamma) + self.lg * np.eye(self.problem.num_random_effects)

    def optimal_tgamma(self, tbeta, gamma):
        tgamma = np.zeros(len(gamma))
        idx = tbeta != 0
        tgamma[idx] = gamma[idx]
        return self.take_only_k_max(tgamma, self.j)

    def good_lambda_gamma(self, mode="upperbound"):
        if mode == "upperbound":
            return sum([np.linalg.norm(self.problem.random_features[i]) ** 4 / np.max(
                self.problem.obs_stds[i] ** 2)
                        for i in range(self.problem.num_groups)])

        elif mode == "exact":
            def minus_max_eig(gamma: np.ndarray):
                self.recalculate_inverse_matrices(gamma)
                A = np.sum([s * s for s in self.zTomegas_invZ], axis=0)
                return -max(np.linalg.eigvals(A))

            gamma_opt = minimize(minus_max_eig, np.ones(self.problem.num_random_effects),
                                 bounds=[(0, None)] * self.problem.num_random_effects).x
            return -minus_max_eig(gamma_opt)

        elif mode == "exact_full_hess":
            def minus_max_eig(gamma: np.ndarray):
                tbeta = np.zeros(self.problem.num_fixed_effects)
                beta = self.optimal_beta_reg(gamma, tbeta)
                return np.min(np.linalg.eigvals(self.hessian_gamma(beta, gamma)))

            gamma_opt = minimize(minus_max_eig, np.ones(self.problem.num_random_effects),
                                 bounds=[(0, 5)] * self.problem.num_random_effects).x
            return -minus_max_eig(gamma_opt), gamma_opt

        else:
            raise Exception("Unknown mode: %s" % mode)



import unittest
from unittest import TestCase

import numpy as np
from numpy import allclose
from scipy.misc import derivative

from skmixed.lme.problems import LinearLMEProblem
from skmixed.lme.oracles import LinearLMEOracleW, LinearLMEOracle


class TestLinearLMEOracleW(TestCase):

    def test_drop_matrices(self):

        problem_parameters = {
            "groups_sizes": [20, 5, 10, 50],
            "features_labels": [1, 2, 3, 3],
            "random_intercept": True,
            "obs_std": 0.1,
            "seed": 42
        }

        problem, _ = LinearLMEProblem.generate(**problem_parameters)
        simple_oracle = LinearLMEOracle(problem)
        oracle = LinearLMEOracleW(problem, lb=0, lg=0,
                                  nnz_tbeta=problem.num_fixed_effects,
                                  nnz_tgamma=problem.num_random_effects)
        trials = 100

        rtol = 1e-10
        atol = 1e-10
        np.random.seed(42)

        for t, (random_beta, random_gamma) in enumerate(zip(np.random.rand(trials, problem.num_fixed_effects),
                                                            np.random.rand(trials, problem.num_random_effects))):
            loss = simple_oracle.loss(random_beta, random_gamma)
            oracle._recalculate_drop_matrices(random_beta, random_gamma)
            w_beta = oracle.drop_penalties_beta
            w_gamma = oracle.drop_penalties_gamma
            for j in range(problem.num_fixed_effects):
                sparse_beta = random_beta.copy()
                sparse_beta[j] = 0
                sparse_gamma = random_gamma.copy()
                idx = oracle.beta_to_gamma_map[j].astype(int)
                if idx >= 0:
                    sparse_gamma[idx] = 0
                    loss3 = simple_oracle.loss(random_beta, sparse_gamma)
                    self.assertTrue(np.isclose(loss3 - loss, w_gamma[idx], rtol=rtol, atol=atol),
                                    msg="%d: W_gamma is not right" % j)
                    loss2 = simple_oracle.loss(sparse_beta, sparse_gamma)
                else:
                    loss2 = simple_oracle.loss(sparse_beta, random_gamma)
                self.assertTrue(np.isclose(loss2 - loss, w_beta[j], rtol=rtol, atol=atol),
                                msg="%d) W_beta is not right" % j)

        sparse_beta = np.zeros(problem.num_fixed_effects)
        sparse_gamma = np.zeros(problem.num_random_effects)
        sparse_beta[0:2] = 1
        sparse_gamma[0] = 1
        oracle._recalculate_drop_matrices(sparse_beta, sparse_gamma)
        w_beta = oracle.drop_penalties_beta
        w_gamma = oracle.drop_penalties_gamma
        self.assertTrue((w_gamma[1:] == 0).all(), msg="Drop of zero gamma is not zero")
        self.assertTrue((w_beta[2:] == 0).all(), msg="Drop of zero beta is not zero")
